#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use Carp 'croak';
use File::Spec;
use List::Util 'max','min','sum';
use Math::BaseCnv 'cnv';

use Math::PlanePath::ComplexMinus;

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyDragon ':all';
our @ipc;
$|=1;

# uncomment this to run the ### lines
# use Smart::Comments;

{
  # ComplexPlus convex hull
  # area 1,2,5,11,25,49,101,205,421,837,1685,3381

  require Math::PlanePath::ComplexPlus;
  require Math::Geometry::Planar;
  my $path = Math::PlanePath::ComplexPlus->new;
  my @values;
  foreach my $k (0 .. 11) {
    print "$k\n";
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my $points = [];
    foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      push @$points,
        [$x,$y],
          [$x+1,$y],
            [$x+1,$y+1],
              [$x,$y+1];

    }
    my $planar = Math::Geometry::Planar->new;
    $planar->points($points);
    if (@$points > 4) {
      $planar = $planar->convexhull2;
      $points = $planar->points;
    }
    my $length = scalar(@$points);
    my $area = $planar->area;

    my $diffs = [ map{ [$points->[$_]->[0] - $points->[$_-1]->[0],
                        $points->[$_]->[1] - $points->[$_-1]->[1] ] }
                  0 .. $#$points
                ];

    print "len=$length  ",points_str($points),"\n";
    print "   ",points_str($diffs),"\n";
    print "   area=$area\n";
    push @values,$area;
  }

  print join(',',@values),"\n";
  shift @values;
  shift @values;
  shift @values;
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}
{
  # MinusNeg, MinusRot, MinusUnrot all
  # 58 states plain
  # more reversed?

  my $f = MinusNegPred_LtoH_flat();
  $f = $f->intersect(MinusRotPred_LtoH_flat());
  $f = $f->intersect(MinusUnrotPred_LtoH_flat());
  $f = $f->MyFLAT::minimize(verbose=>1);
  $f = $f->MyFLAT::reverse
    ->MyFLAT::minimize(verbose=>1);
  $f->MyFLAT::view;
  exit 0;
}

{
  # MinusRot bit patterns to each state
  my $f = MinusRotIncPred_dir8_LtoH_flat(0);
  # $f = $f->MyFLAT::reverse
  #   ->MyFLAT::minimize(verbose=>1);
  $f->MyFLAT::view;
  exit 0;
}
{
  # MinusConj states
  my @pending = ([ [1,0], [0,0] ]);
  my %fdz_str_to_state;
  my %seen;
  my $state_upto = 0;
  my @state_table;
  my $fdz_to_str = sub {
    my ($fdz) = @_;
    my ($fx,$fy) = @{$fdz->[0]};
    my ($dzx,$dzy) = @{$fdz->[1]};
    my $fdz_str = "$fx,$fy $dzx,$dzy";
    return "$fx,$fy $dzx,$dzy";
  };
  my @state_str;
  my $get_state = sub {
    my ($fdz) = @_;
    my $str = $fdz_to_str->($fdz);
    if (! exists $fdz_str_to_state{$str}) {
      $fdz_str_to_state{$str} = $state_upto;
      $state_str[$state_upto] = "s$state_upto $str";
      $state_upto++;
    }
    return $fdz_str_to_state{$str};
  };
  my @accepting;
  while (@pending) {
    my $fdz = shift @pending;
    my $state = $get_state->($fdz);
    my ($fx,$fy) = @{$fdz->[0]};
    my ($dzx,$dzy) = @{$fdz->[1]};
    print "f=$fx,$fy dz $dzx,$dzy state $state\n";
    foreach my $bit (0,1) {
      my $out = ($bit + $dzx + $dzy) % 2;
      my ($num_x,$num_y) = ($fx*$bit - $out + $dzx,
                            $fy*$bit        + $dzy);
      my ($next_dzx,$next_dzy) = i_minus_1_div($num_x, $num_y);
      my ($next_fx,$next_fy) = xy_rotate_plus90($fx,$fy);
      my $next_fdz = [ [$next_fx,$next_fy], [$next_dzx,$next_dzy] ];
      my $next_str = $fdz_to_str->($next_fdz);
      if (! $seen{$next_str}++) {
        push @pending, $next_fdz;
      }
      my $next_state = $get_state->($next_fdz);
      $state_table[$state]->[$bit] = $next_state;
      print "  bit=$bit out=$out to $next_state  $next_fx,$next_fy $next_dzx,$next_dzy     num $num_x,$num_y\n";
      is_integer($next_dzx) or die;
      is_integer($next_dzy) or die;
    }
  }

  require MyFLAT;
  {
    my @f_initial_state;
    {
      my ($x,$y) = (1,0);
      foreach (0..3) {
        push @f_initial_state, $get_state->([ [$x,$y], [0,0] ]);
        ($x,$y) = xy_rotate_plus90($x,$y);
      }
    }
    foreach my $i (0..3) {
      my $state = $f_initial_state[$i];
      print "initial $i $state  $state_str[$state]\n";
    }

    # ### as nfa ...
    # $fa = $fa->MyFLAT::as_nfa($fa);
    # ### reverse ...
    # $fa = $fa->reverse;
    # ### as_dfa: $fa->num_states
    # $fa = $fa->MyFLAT::minimize(verbose=>1);
    #
    # $fa = $fa->as_dfa;
    # return $fa;

    my @f_list = map {
      ### $_
      MyFLAT::aref_to_FLAT_DFA (\@state_table,
                                starting => $f_initial_state[$_],
                                accepting_list => \@f_initial_state)
          ->MyFLAT::as_nfa
          ->MyFLAT::set_name("starting $_")
          ->MyFLAT::minimize(verbose=>1);
    } 0..3;
    foreach my $i (0..3) {
      my $f = $f_list[$i];
      print $f->{'name'}," ",$f->num_states," starting ",
        join(',',$f->get_starting),"\n";
      $f_list[$i]->MyFLAT::view;
    }

    my @mod_list = map {
      my $mod = ($_)%4;
      FLAT::Regex->new(("(0|1)" x ($mod))
                       . "((0|1)(0|1)(0|1)(0|1))*")
          ->as_nfa
          ->MyFLAT::set_name("len $mod mod 4")
          ->MyFLAT::minimize(verbose=>1);
    } 0..3;

    foreach my $i (0..3) {
      $f_list[$i] = $f_list[$i]->intersect($mod_list[$i])
        ->MyFLAT::set_name("int$i")
        ->MyFLAT::minimize(verbose=>1);
    }

    foreach my $i (0..3) {
      ### reverse ...
      $f_list[$i] = $f_list[$i]->MyFLAT::reverse($f_list[$i])
        ->MyFLAT::minimize(verbose=>1);
    }

    my $f = $f_list[0]
      ->MyFLAT::minimize(verbose=>1);

    foreach my $i (1..1) {
      $f = $f->union($f_list[$i])
        ->MyFLAT::set_name("u$i")
        ->MyFLAT::minimize(verbose=>1);
    }
    $f->MyFLAT::view;
    exit 0;
  }
  {
    my $f = MyFLAT::aref_to_FLAT_DFA(\@state_table,
                                     accepting_list => [0],
                                    );
    MyFLAT::FLAT_print_tikz($f, state_labels => \@state_str);
    $f = $f->MyFLAT::blocks(4);
    $f = $f->MyFLAT::minimize(verbose=>1);
    MyFLAT::FLAT_print_tikz($f);
    $f->MyFLAT::view;
    exit 0;
  }
  exit 0;
}
{
  # MinusNeg etc singletons
  my $type = 'MinusRot';
  $type = 'MinusDouble';
  my $max_k = to_bigint(50);
  {
    my @a = MyFLAT::FLAT_count_contains(NonNSEWpred($type),$max_k);
    print "NonNSEWpred  [",join(', ',@a),"]\n";
  }
  {
    my $f = NonNSEWpred($type);
    foreach my $non_dir8 (1,3,5,7) {
      $f = $f->intersect(NeighbourPred($type,$non_dir8)->complement)
        ->MyFLAT::minimize(verbose=>1);
    }
    my @a = MyFLAT::FLAT_count_contains($f,$max_k);
    print "single8      [",join(', ',@a),"]\n";
  }

  exit 0;
}
{
  # MinusNeg etc hanging squares

  # recurrence_guess()

  my $type = 'MinusRot';
  my $max_k = to_bigint(50);
  my @H = (0) x ($max_k+1);
  foreach my $dir8 (1,3,5,7) {
    my $f = HangingSquarePred_dir8($type,$dir8);
    my @a = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $H[$i] += $a[$i]; }
    print "dir8=$dir8  ",join(', ',@a),"\n";
  }
  print "H [",join(', ',@H),"]\n";
  exit 0;
}


{
  # MinusDoubleB boundary length

  # recurrence_guess([4, 4, 4, 6, 10, 16, 26, 46, 80, 136, 230, 394, 676, 1148, 1954, 3334, 5668, 9632, 16378, 27818, 47232, 80204, 136134, 231006, 392000, 665080, 1128206, 1913786, 3246156, 5505612, 9337466, 15835782, 26855308, 45541856, 77229714, 130962938, 222078184, 376581820, 638569214, 1082811262, 1836094856, 3113400536, 5279255942, 8951771242, 15179026772, 25738171676, 43642598226, 74001886278, 125479950132, 212767547744, 360774673386, 611739249162, 1037280867472, 1758839321580, 2982330524022, 5056909980126, 8574613360784, 14539308924408, 24653177031102, 41802470948826, 70881182566300, 120187667436972, 203792791873386, 345555411739462, 585931101999004, 993517181631776, 1684628697116418, 2856491866619098, 4843527577040248, 8212786851170524, 13925773207423790, 23612832021324158, 40038410830617688, 67889964371392856, 115115638356695990, 195192473891656202, 330972421993249572, 561203725717878780, 951588711191295682, 1613533607769134982, 2735941132586174660, 4639118657360212256, 7866186015771267034, 13338068480297520746, 22616306073182119840, 38348678492861669772, 65024815995037304870, 110257428897083798110, 186954786937245343968, 317004420398618156792, 537519280245730853742, 911428856984780207866, 1545437701779051133932, 2620476267847701245964, 4443333989599324094810, 7534209404016054253830, 12775161954862868167340, 21661829955202829627744, 36730248792734256350834, 62280572743621443419834, 105604232711461140638472])

  # initial exceptions [4, 4]
  # 1/20*vector_modulo([0, 1, -1, 1, 0, -1, 1, -1],n) * 2^floor(n/2)
  # + vector_modulo([89/34, 107/85, 96/85, 283/170, 81/34, 97/85, -79/85, 193/170],n)
  # + gf 1/68*(60 + 21*x + 53*x^2)/(1 - x - 2*x^3)
  # + gf (3/2 - x - 7/2*x^3)/(1 - 2*x^3 - x^4)
  #
  # 1.39533699446707301879314

  # gf_in_terms_of(1/68*(60 + 21*x + 53*x^2)/(1 - x - 2*x^3), \
  #                gComplexBoundary(x))

  my $max_k = to_bigint(20);

  my @MinusDoubleA = MyFLAT::FLAT_count_contains(MinusDoublePred_LtoH_flat(),$max_k);
  print "Within  ",join(', ',@MinusDoubleA),"\n";
  
  my $type = 'MinusDouble';
  my @MinusDoubleB = (0) x ($max_k+1);
  foreach my $dir8 (0,4,2,6) {
    my $f = BoundaryPred($type, $dir8);
    my @a = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $MinusDoubleB[$i] += $a[$i]; }
    print "dir8=$dir8  ",join(', ',@a),"\n";
  }
  print "B      [",join(', ',@MinusDoubleB),"]\n";

  exit 0;
}

{
  # MinusNegB boundary length

  # k=4 N=3 0011 at z=I has east as N=14 1110 z=1+I
  # MinusPoint(3)
  # MinusPoint(14)
  # East 0100  N=2        West 1000  N=1
  #      0110  N=6             1110  N=7
  #      1100  N=3             1111  N=15
  #
  #             I  1+I      [2] 3  14 [15]
  # k=4         0               0
  #       -1-I -I            6  7

  # B =
  # + gf 1/(1 - x + 2*x^2)                          # sqrt2
  # + gf (8/3 + 4/3*x + 8/3*x^2)/(1 - x - 2*x^3)    # dragon (1 - x - 2*x^3)
  # + gf (-1 - x)/(1 - x^2 - 2*x^3)
  # + gf (4/3 + 4/3*x + 2/3*x^2 + 8/3*x^3)/(1 + x + x^2 + x^3 + 2*x^4)
  #
  # (1 - x^2 - 2*x^3) 1.52137970680456756960408
  # (1 + x + x^2 + x^3 + 2*x^4)  1.25197139334208603724903

  my $NegNonB_east = MinusNegPred_LtoH_flat()
    ->intersect(MinusNonB_east_LtoH_flat())
    ->intersect(MinusNeg_west_LtoH_flat())
    ->MyFLAT::minimize
    ->MyFLAT::set_name('E');

  my $NegNonB_west = MinusNegPred_LtoH_flat()
    ->intersect(MinusNonB_west_LtoH_flat())
    ->intersect(MinusNeg_east_LtoH_flat())
    ->MyFLAT::minimize
    ->MyFLAT::set_name('NegNonB_west');

  my $NegNonB_north = MinusNegPred_LtoH_flat()
    ->intersect(MinusNonB_north_LtoH_flat())
    ->intersect(MinusNeg_south_LtoH_flat())
    ->MyFLAT::minimize
    ->MyFLAT::set_name('N');
  my $NegNonB_south = MinusNegPred_LtoH_flat()
    ->intersect(MinusNonB_south_LtoH_flat())
    ->intersect(MinusNeg_north_LtoH_flat())
    ->MyFLAT::minimize->MyFLAT::set_name('S');

  MyFLAT::FLAT_show_breadth($NegNonB_east, 4, 'lowtohigh');
  MyFLAT::FLAT_show_breadth($NegNonB_west, 4, 'lowtohigh');
  MyFLAT::FLAT_show_breadth($NegNonB_north, 4, 'lowtohigh');
  MyFLAT::FLAT_show_breadth($NegNonB_south, 4, 'lowtohigh');

  my $NegB_east = MinusNegPred_LtoH_flat()->difference($NegNonB_east);
  my $NegB_west = MinusNegPred_LtoH_flat()->difference($NegNonB_west);
  my $NegB_north = MinusNegPred_LtoH_flat()->difference($NegNonB_north);
  my $NegB_south = MinusNegPred_LtoH_flat()->difference($NegNonB_south);

  my $max_k = to_bigint(120);
  my @MinusNegA = MyFLAT::FLAT_count_contains(MinusNegPred_LtoH_flat(),$max_k);
  print "Within ",join(', ',@MinusNegA),"\n";

  my @NegB_east_counts = MyFLAT::FLAT_count_contains($NegB_east,$max_k);
  my @NegB_west_counts = MyFLAT::FLAT_count_contains($NegB_west,$max_k);
  my @NegB_north_counts = MyFLAT::FLAT_count_contains($NegB_north,$max_k);
  my @NegB_south_counts = MyFLAT::FLAT_count_contains($NegB_south,$max_k);
  print "E  ",join(', ',@NegB_east_counts),"\n";
  print "W  ",join(', ',@NegB_west_counts),"\n";
  print "N  ",join(', ',@NegB_north_counts),"\n";
  print "S  ",join(', ',@NegB_south_counts),"\n";

  my @MinusNegB
    = map { $NegB_east_counts[$_]
            + $NegB_west_counts[$_]
            + $NegB_north_counts[$_]
            + $NegB_south_counts[$_]
          } 0 .. $max_k;
  print "B  ",join(', ',@MinusNegB),"\n";

  # my @MinusNegB = map { 4*$MinusNegA[$_] - $MinusNegNonB[$_]
  #                     } 0 .. $#MinusNegNonB;
  # print "B      [",join(', ',@MinusNegB),"]\n";

  exit 0;
}
{
  # MinusPoint(7)
  # MinusInc(7,1) == 58
  # MinusInc(7,I) == 0
  # MinusInc(7,1+I) == 1
  # MinusDoublePred(3,7)
  # MinusDoublePred(3,58)
  # MinusDoublePred(3,0)
  # MinusDoublePred(8,1)
  # 000  N=0
  # 111  N=7
  # my(k=8); for(n=0,2^k-1, \
  #   if(MinusDoublePred(k,n) \
  #      && !MinusDoublePred(k,MinusInc(n,1  )) \
  #      && !MinusDoublePred(k,MinusInc(n,  I)) \
  #      && !MinusDoublePred(k,MinusInc(n,1+I)), \
  #      print(n,"  ",MinusPoint(n))));

  # my(k=3); for(n=0,2^k-1, \
  #   print(n,"  ",MinusPoint(n), \
  #         " ",MinusDoublePred(k,n) \
  #         " ",MinusDoublePred(k,MinusInc(n,1  )) \
  #         " ",MinusDoublePred(k,MinusInc(n,  I)) \
  #         " ",MinusDoublePred(k,MinusInc(n,1+I))));

  #                [3 at I]
  # k=3             0 at 0   [1 at 2]
  #    [6 at -1-I]  7 at -I
  #                [4 at -2I]
  
  my $dir8 = 0;
  my $dir8_plus1 = ($dir8+1) % 8;
  my $dir8_plus2 = ($dir8+2) % 8;

  MyFLAT::FLAT_show_breadth(MinusDoublePred_LtoH_flat(), 3, 'lowtohigh');

  my $f_not = MinusDoubleNeighbourPred_dir8_LtoH_flat($dir8);
  MyFLAT::FLAT_show_breadth($f_not, 3, 'lowtohigh');

  my $f_not_plus1 = MinusDoubleNeighbourPred_dir8_LtoH_flat($dir8_plus1);
  MyFLAT::FLAT_show_breadth($f_not_plus1, 3, 'lowtohigh');

  my $f_not_plus2 = MinusDoubleNeighbourPred_dir8_LtoH_flat($dir8_plus2);
  MyFLAT::FLAT_show_breadth($f_not_plus2, 3, 'lowtohigh');

  my $f = MinusDoublePred_LtoH_flat()
    ->intersect($f_not_plus1->complement)
    ->intersect($f_not_plus1->complement)
    ->intersect($f_not_plus2->complement);
  MyFLAT::FLAT_show_breadth($f, 3, 'lowtohigh');
  exit 0;
}

{
  # MinusDouble pieces Eight by path

  # recurrence_guess([0,1,1,1, 1,1,1,2, 3,4,5,7, 11,15,21])

  my $path = Math::PlanePath::ComplexMinus->new;
  my $prev = 0;
  my @values;
 K: foreach my $k (0..14) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my @n_to_piece;
    my $upto_piece = 0;
    my %pieces;
  N: foreach my $n (0 .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      next unless $path->xy_to_n(2*$x,2*$y) <= $n_hi;
      my $piece = $upto_piece++;
      $n_to_piece[$n] = $piece;
      ### at: "n=$n  xy=$x,$y  this piece=$piece"

      $pieces{$piece} = 1;
      foreach my $dir8 (0 .. 7) {
        my $dx = $dir8_to_dx[$dir8];
        my $dy = $dir8_to_dy[$dir8];
        my $x2 = $x+$dx;
        my $y2 = $y+$dy;
        my $n2 = $path->xy_to_n($x2,$y2);
        next unless $n2 <= $n_hi;
        next unless $path->xy_to_n(2*$x2,2*$y2) <= $n_hi;
        my $p = $n_to_piece[$n2];
        if (defined $p && $p != $piece) {
          delete $pieces{$p};
          foreach (@n_to_piece) {
            if (defined $_ && $_ == $p) { $_ = $piece; }  # mutate array
          }
          ### unify: "$p, now ".scalar keys %pieces
        }
      }
    }
    my $num_pieces = scalar keys %pieces;
    my $diff = $num_pieces - $prev;
    print "k=$k  $num_pieces  +$diff\n";
    $prev = $num_pieces;
    push @values, $num_pieces;
  }
  print "pieces [",join(',',@values),"]\n";
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}



{
  # MinusRotB boundary length

  # recurrence_guess([4, 4, 6, 12, 18, 32, 58, 94, 168, 294, 484, 852, 1466, 2444, 4242, 7238, 12172, 20914, 35540, 60060, 102566, 174000, 294738, 501614, 850480, 1441982, 2449620, 4152804, 7043074, 11953332, 20264994, 34370006, 58303972, 98851306, 167645956, 284321260, 482075342, 817521368, 1386330178, 2350636222, 3986104168, 6759177590, 11460920228, 19434280308, 32953674058, 55876927964, 94748561010, 160658566502, 272416620668, 461921838498, 783245916916, 1328091449148, 2251956258230, 3818466657664, 6474684999026, 10978652317966, 18615636312288, 31565107014830, 53522553277748, 90753966646788, 153884462854226, 260929935273348, 442438264271874, 750207970675958, 1272068788848212, 2156946437938394, 3657364514988708, 6201504560477324, 10515400617897822, 17830135433749928, 30233151031076066, 51263961291556446, 86924247707387032, 147390566924402790, 249918515015778180, 423767051948168596, 718548231692486426, 1218385333444228524, 2065919547694079698, 3503016135066331782, 5939787002350812524, 10071626390237426066, 17077658998245420052, 28957233559168013276, 49100487114118138310, 83255806037881368912, 141170274689870687378, 239371250970001518254, 405882865603748520080, 688223419185717643934, 1166965926573273136276, 1978731664859838638628, 3355178514680047028130, 5689110382336497297492, 9646573731674164659426, 16356930792072334680086, 27735151595552719702980, 47028299113258963260874, 79742160781208396527940, 135212464076582850120492, 229269062453509246361326])

  # initial exceptions [4]
  # vector_modulo([24/17, 12/17, 23/17, 20/17, 10/17, 22/17, 11/17, 14/17],n)
  # + gf (4/7 + 3/7*x - 2/7*x^2)/(1 + x + x^2 - x^3)
  # + gf (359/119 + 169/119*x + 348/119*x^2)/(1 - x - 2*x^3)  \\ dragon
  # + gf (-2 - x - 2*x^2)/(1 - 2*x^3 - x^4)

  # abs(apply(z->1/z,polroots(1 + x + x^2 - x^3)))
  # abs(apply(z->1/z,polroots(1 - x - 2*x^3)))
  # abs(apply(z->1/z,polroots(1 - 2*x^3 - x^4)))
  # not in OEIS: 1.35620306562629518801434
  # not in OEIS: 1.39533699446707301879314

  my $max_k = to_bigint(100);

  my @MinusRotA = MyFLAT::FLAT_count_contains(MinusRotPred_LtoH_flat(),$max_k);
  print "Within  ",join(', ',@MinusRotA),"\n";

  my @MinusRotB = (0) x ($max_k+1);
  foreach my $dir8 (0,4,2,6) {
    my $f = MinusRotPred_LtoH_flat()
      ->intersect(MinusRotNeighbourPred_dir8_LtoH_flat($dir8)->complement)
      ->MyFLAT::minimize
      ->MyFLAT::set_name("MinusRotB dir8=$dir8");

    my @a = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $MinusRotB[$i] += $a[$i]; }
    print "dir8=$dir8  ",join(', ',@a),"\n";
  }
  print "B      [",join(', ',@MinusRotB),"]\n";

  exit 0;
}

{
  # MinusRot pieces Eight

  # recurrence_guess([1,1,1,1,1,1,3,4,5,11,10,16,31,25,50,86,71,148,229,195,426,600,548,1201,1557,1558,3332,4023,4464,9126,10412,12827,24738,27113,36788,66469,71265,104968,177341,189437,297464,470600,509477,836449,1244304,1385129,2333281,3284257,3800074,6457928,8669290,10496229,17742890,22925265,29117899,48424107,60823979,80944609,131394160,162091083,225049727,354805477,434186318,624844581,954469253,1169325550,1730573575,2560746182,3165606658,4777721456,6859294036,8609969016,13143060779,18363516964,23508023010,36022103474,49182498368,64369937434,98375464816,131885894634,176590718480,267776962331,354321608078,484905227656,726788488425,954112791208,1331633501525,1967931999012,2575755451006,3654690057806,5318895650427,6971475866844,10019014713552,14357967396869,18914515164877,27425230622000,38732042436776,51426847249938,74943759284220,104467776026472,140068343053794,204432082045753,281857449523659,381991213837426,556685333482913,760985541238687,1042635242418750,1513480781775110,2056580709913903,2847008196371049,4109039242692450,5564371888538292,7774177149292115,11143219384908048,15073799956916305,21222156593459369,30193301278776298,40884720469883104,57900957019361468,81764699464011295,111017859201159313,157858421088516442,221359392697856296,301754711518564205,430023157555396192,599263493405263110,820848366288981036,1170423645331530297,1622634262030786574,2234233040170183765,3182961343802392720,4395237786020658059,6083561853080207105,8649414287288712800,11911262462248466981,16567738896971623659,23488446035976066900,32298331745081419452,45119669519368058937,63751364780770074455,87631838984412016992,122857136776874453913,172962937441424497201,237901767854294013823,334438315066001646806,469144622387699110436,646199384998896046262,910076077937278846027,1272358578769510469930,1756047605355282583673,2475513988926533294083,3450757593381539451801,4773831196491547796455,6730896266344716902661,9359795443447011973748,12981210442482339697153,18293887688442244904239,25392199005617172671461,35305030972889700896878,49702723281771651458115,68903445445172190587313,96026178387974161701562,134995137816764638189251,187026196200702411360556,261178477348780088122978,366561857145589119596008,507796033995010583989852,710311919124995357143782,995170322760470116214343,1379105529565206852218744,1931528904921625714272470,2701460058660226397833896,3746416459629565161038544,5251429800332110548110762,7332954977594035710468064,10179555639382428592422984,14274773904402409451567516,19905085205300444102755415,27664085623933079439810234,38794836474941422011660910,54034937370455198730187945,75189549738534713937659760,105413205259750085950841645,146698964300001447102153106,204376630362808947442965626,286378560344535099859478354,398320193342118072927670879,555543158972038356032884370,777896357709595254364734784,1081674465907001217239730853,1510079319696667720310516621,2112770153760242065123178830,2937793720622341524691205524,4104523203243052556970007358,5737815377544705315961447064,7980020177833399585838783054,11155671891895134601002697730,15581855361044854220592803793,21678985148540899964900826739,30317364908368570241253940320,42313953654622937143216632445])

  # initial exceptions [1, 1, 1, 1, 1]
  # -1/4 * n
  # + vector_modulo([-1/2, 3/2, 7/8, 0, 9/8, 3/8, -1/4, 7/8, 3/2, 1/2, 7/8, 3/2, 1/8, -5/8, 7/4, 3/8, 1/2, 3/2, -1/8, 0, 9/8, -5/8, 3/4, 7/8, -1/2, 3/2, 7/8, -1/2, 9/8, 3/8, -1/4, 3/8, 3/2, 1/2, 7/8, 1, 1/8, -5/8, 7/4, -1/8, 1/2, 3/2, -1/8, 1/2, 9/8, -5/8, 3/4, 11/8],n)
  # + gf (33/16 + 25/8*x + 53/16*x^2)/(1 + x + x^2 - x^3)
  # + gf (-89/16 + 59/16*x - 9/16*x^2 + 209/16*x^3)/(1 - 2*x^3 - x^4)

  my $max_k = to_bigint(200);
  my @net_turn = (0) x $max_k;

  # neither a,b nor diagonal,  turn +1
  foreach my $dir8 (0,4,2,6) {
    my $dir8_plus1 = ($dir8+1) % 8;
    my $dir8_plus2 = ($dir8+2) % 8;
    my $f = MinusRotPred_LtoH_flat()
      ->intersect(MinusRotNeighbourPred_dir8_LtoH_flat($dir8)->complement)
      ->intersect(MinusRotNeighbourPred_dir8_LtoH_flat($dir8_plus1)->complement)
      ->intersect(MinusRotNeighbourPred_dir8_LtoH_flat($dir8_plus2)->complement)
      ;
    my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $net_turn[$i] += $counts[$i]; }
    print "dir8=$dir8 turn+1 [",join(',',@counts),"]\n";
    # if ($dir8==0) {
    #   MyFLAT::FLAT_show_breadth($f, 5, 'lowtohigh');
    # }
  }

  # with NE, turn -1
  foreach my $dir8 (0,4,2,6) {
    my $dir8_plus1 = ($dir8+1) % 8;
    my $dir8_plus2 = ($dir8+2) % 8;
    my $f = MinusRotPred_LtoH_flat()
      ->intersect(MinusRotNeighbourPred_dir8_LtoH_flat($dir8)->complement)
      ->intersect(MinusRotNeighbourPred_dir8_LtoH_flat($dir8_plus1))
      ->intersect(MinusRotNeighbourPred_dir8_LtoH_flat($dir8_plus2)->complement)
      ;
    my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $net_turn[$i] -= $counts[$i]; }
    print "dir8=$dir8 2*turn-1 [",join(',',@counts),"]\n";
  }

  # both a,b  and no diagonal turn -1
  foreach my $dir8 (0,4,2,6) {
    my $dir8_plus1 = ($dir8+1) % 8;
    my $dir8_plus2 = ($dir8+2) % 8;
    my $f = MinusRotPred_LtoH_flat()
      ->intersect(MinusRotNeighbourPred_dir8_LtoH_flat($dir8))
      ->intersect(MinusRotNeighbourPred_dir8_LtoH_flat($dir8_plus1)->complement)
      ->intersect(MinusRotNeighbourPred_dir8_LtoH_flat($dir8_plus2))
      ;
    my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $net_turn[$i] -= $counts[$i]; }
    print "dir8=$dir8 both    [",join(',',@counts),"]\n";
    # if ($dir8==0) {
    #   MyFLAT::FLAT_show_breadth($f, 5, 'lowtohigh');
    # }
  }

  @net_turn = map {$_/4} @net_turn;
  print "net [",join(',',@net_turn),"]\n";
  exit 0;
}


{
  # MinusNeg pieces Eight by state machine corner turns, diagonal included

  # recurrence_guess([1,1,1,1,1,3,7,7,7,9,21,45,61,63,83,163,323,485,577,745,1305,2427,3807,4991,6591,10625,18733,29925,41781,56951,87595,147355,236411,343773,482521,725537,1176209,1878835,2803255,4030711,6008439,9485817,15023429,22759069,33344301,49642415,77000451,120796051,184474291,274055893,408947313])

  #   1/3*  1/(1 - x)
  # + 3/16* 1/(1 + x)
  # + (1/16 + 3/8*x)/(1 - x + 2*x^2)
  # + 1/2*(1 + x - 1/2*x^2)/(1 - x^2 - 2*x^3)
  # + (-1/12 - 1/6*x - 1/4*x^2 - 1/3*x^3)/(1 + x + x^2 + x^3 + 2*x^4)

  my $max_k = to_bigint(50);
  my @net_turn = (0) x $max_k;

  # neither a,b nor diagonal,  turn +1
  foreach my $dir8 (0,4,2,6) {
    my $dir8_plus1 = ($dir8+1) % 8;
    my $dir8_plus2 = ($dir8+2) % 8;
    my $f = MinusNegPred_LtoH_flat()
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8)->complement)
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8_plus1)->complement)
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8_plus2)->complement)
      ;
    my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $net_turn[$i] += $counts[$i]; }
    print "dir8=$dir8 turn+1 [",join(',',@counts),"]\n";
    # if ($dir8==0) {
    #   MyFLAT::FLAT_show_breadth($f, 5, 'lowtohigh');
    # }
  }

  # neither a,b but with diagonal,  2* turn -1
  # but this is counted on the opposing diagonal point too, so just 1*
  foreach my $dir8 (0,4,2,6) {
    my $dir8_plus1 = ($dir8+1) % 8;
    my $dir8_plus2 = ($dir8+2) % 8;
    my $f = MinusNegPred_LtoH_flat()
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8)->complement)
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8_plus1))
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8_plus2)->complement)
      ;
    my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $net_turn[$i] -= $counts[$i]; }
    print "dir8=$dir8 2*turn-1 [",join(',',@counts),"]\n";
    # if ($dir8==0) {
    #   MyFLAT::FLAT_show_breadth($f, 5, 'lowtohigh');
    # }
  }

  # both a,b  and no diagonal turn -1
  foreach my $dir8 (0,4,2,6) {
    my $dir8_plus1 = ($dir8+1) % 8;
    my $dir8_plus2 = ($dir8+2) % 8;
    my $f = MinusNegPred_LtoH_flat()
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8))
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8_plus1)->complement)
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8_plus2))
      ;
    my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $net_turn[$i] -= $counts[$i]; }
    print "dir8=$dir8 both    [",join(',',@counts),"]\n";
    # if ($dir8==0) {
    #   MyFLAT::FLAT_show_breadth($f, 5, 'lowtohigh');
    # }
  }

  @net_turn = map {$_/4} @net_turn;
  print "net [",join(',',@net_turn),"]\n";
  exit 0;
}
{
  # MinusNeg pieces Eight by path

  # recurrence_guess([0,1,1,1,1,3,7,7,7,9,21,45,61,63,83])

  my $path = Math::PlanePath::ComplexMinus->new;
  my $prev = 0;
  my @values;
 K: foreach my $k (0..14) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my @n_to_piece;
    my $upto_piece = 0;
    my %pieces;
  N: foreach my $n (0 .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      next unless $path->xy_to_n(-$x,-$y) <= $n_hi;
      my $piece = $upto_piece++;
      $n_to_piece[$n] = $piece;
      ### at: "n=$n  xy=$x,$y  this piece=$piece"

      $pieces{$piece} = 1;
      foreach my $dir8 (0 .. 7) {
        my $dx = $dir8_to_dx[$dir8];
        my $dy = $dir8_to_dy[$dir8];
        my $x2 = $x+$dx;
        my $y2 = $y+$dy;
        my $n2 = $path->xy_to_n($x2,$y2);
        next unless $n2 <= $n_hi;
        next unless $path->xy_to_n(-$x2,-$y2) <= $n_hi;
        my $p = $n_to_piece[$n2];
        if (defined $p && $p != $piece) {
          delete $pieces{$p};
          foreach (@n_to_piece) {
            if (defined $_ && $_ == $p) { $_ = $piece; }  # mutate array
          }
          ### unify: "$p, now ".scalar keys %pieces
        }
      }
    }
    my $num_pieces = scalar keys %pieces;
    my $diff = $num_pieces - $prev;
    print "k=$k  $num_pieces  +$diff\n";
    $prev = $num_pieces;
    push @values, $num_pieces;
  }
  print "pieces [",join(',',@values),"]\n";
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}



{
  # MinusNeg singletons no neighbours NSEW direction, by state
  # 1, 1, 1, 0, 0, 4, 6, 4, 4, 6, 24, 40, 36, 36, 66, 168, 292
  # 0  1  2  3  4  5  6  7  8
  # smaller than dragon cubic
  #
  # initial exceptions [1]
  # vector_modulo([7/24, 1/24],n)
  # + gf (5/16 - 1/16*x)/(1 - x + 2*x^2)
  # + gf (7/16 + 1/8*x + 5/16*x^2)/(1 - x^2 - 2*x^3)
  # + gf (5/24 + 19/24*x + 9/8*x^2 + 23/24*x^3)/(1 + x + x^2 + x^3 + 2*x^4)
  #
  # polroots(x^3 - x^2 - 2)  dragon
  # polroots(x^3 - x - 2)   1.52 < 1.69

  # recurrence_guess([1,1,1,0,0,4,6,4,4,6,24,40,36,36,66,168,292,326,348,572,1244,2180,2798,3260,4900])

  my $f = MinusNegPred_LtoH_flat();
  foreach my $dir8 (0,2,4,6) {
    $f = $f->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8)->complement)
      ->MyFLAT::minimize(verbose=>1);
  }
  $f->MyFLAT::set_name('MinusNeg singleton Four');

  my $max_k = to_bigint(100);
  my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
  print "singletons [",join(',',@counts),"]\n";

  # $f->MyFLAT::view;
  exit 0;
}
{
  # MinusNeg singletons no neighbours 8-sides,  by state

  # recurrence_guess([1,1,1,0,0,2,6,4,2,2,12,32,36,22,30,88,210,286,252,300,684,1482,2230,2428,2898,5418,10908,17288,21476,26734,44158,82608,134338,181670,237036,366132,641356,1049730,1500790,2041076,3053986,5079314,8265964,12242480,17221732,25459846,40797598,65625928,99253874,143316974,211526044])

  # initial exceptions [1]
  # vector_modulo([7/12, 1/12],n)
  # + gf (3/16 + 3/16*x)/(1 - x + 2*x^2)
  # + gf (7/16 + 1/8*x - 3/16*x^2)/(1 - x^2 - 2*x^3)
  # + gf (1/24 + 11/24*x + 5/8*x^2 + 7/24*x^3)/(1 + x + x^2 + x^3 + 2*x^4)

  my $f = MinusNegPred_LtoH_flat();
  foreach my $dir8 (0 .. 7) {
    $f = $f->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8)->complement)
      ->MyFLAT::minimize(verbose=>1);
  }
  $f->MyFLAT::set_name('MinusNeg singleton Eight');

  my $max_k = to_bigint(50);
  my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
  print "singletons [",join(',',@counts),"]\n";

  # $f->MyFLAT::view;
  exit 0;
}

{
  # MinusNeg pieces Four by state machine corner turns

  # recurrence_guess([1,1,1,1,1,5,7,7,9,13,33,53,61,77,119,243,405,525,673,1017,1865,3125,4375,5823,8593,14669,24369,35773,49445,72253,117415,192043,289853,413325,604049,950593,1528081,2340389,3417319,5019895,7746009,12256397,18889601,28052037,41495117,63317421,98903447,152631523,229264037,341509133,518144193,801527945,1235392537,1869093013,2801252471,4240109487,6514118369,10016750541,15219182353,22921554381,34679465397,53033760925,81342621511,123865198363,187242639053,283437539469,432195761201,661370742737,1008073516641,1527869643141,2314915639751,3523904399591,5382339963561,8205645122829,12458690622433,18895043405909,28737620598109,43830624995597,66810625508791,101552393199315,154152551983797])

  # 1/12/(1 - x)
  # + 1/4* 1/(1 - x + 2*x^2)                  \\ sqrt7i
  # + 1/2* (1 + x + x^2)/(1 - x^2 - 2*x^3)    \\ 1.521
  # + 1/6* (1 + 2*x + 3*x^2 + 4*x^3)/(1 + x + x^2 + x^3 + 2*x^4)

  # +---+---+
  # | b |   |    neither a,b  turn +1
  # +---+---+    one of a,b   turn  0
  # | X | a |    both a,b     turn -1
  # +---+---+

  #                 3  14      1+1+2+2=6 turn +1
  # k=4             0          0+0+1+1=2 turn -1
  #              6  7

  #              25           9     6+6+6+6=24 turn +1      5,9,13,14,21,25
  #                  3 14           1+1+1+1= 4 turn -1
  # k=5    17 28 29  0  1 12 13            ---
  #               6  7                      20/4 = 5 pieces
  #        21           5
  # my(k=5); select(n->MinusNegPred(k,n),vector(2^k,n,n--; n))

  my $max_k = to_bigint(80);
  my @net_turn = (0) x $max_k;

  # neither a,b  turn +1
  foreach my $dir8 (0,4,2,6) {
    my $dir8_plus1 = ($dir8+1) % 8;
    my $dir8_plus2 = ($dir8+2) % 8;
    my $f = MinusNegPred_LtoH_flat()
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8)->complement)
      # ->MyFLAT::minimize(verbose=>1)
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8_plus2)->complement)
      # ->MyFLAT::minimize(verbose=>1)
      ;
    my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $net_turn[$i] += $counts[$i]; }
    print "dir8=$dir8 neither [",join(',',@counts),"]\n";
    if ($dir8==0) {
      MyFLAT::FLAT_show_breadth($f, 5, 'lowtohigh');
    }
  }

  # both a,b  turn -1
  foreach my $dir8 (0,4,2,6) {
    my $dir8_plus1 = ($dir8+1) % 8;
    my $dir8_plus2 = ($dir8+2) % 8;
    my $f = MinusNegPred_LtoH_flat()
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8))
      # ->MyFLAT::minimize(verbose=>1)
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8_plus1)->complement)
      # ->MyFLAT::minimize(verbose=>1)
      ->intersect(MinusNegNeighbourPred_dir8_LtoH_flat($dir8_plus2))
      # ->MyFLAT::minimize(verbose=>1)
      ;
    my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $net_turn[$i] -= $counts[$i]; }
    print "dir8=$dir8 both    [",join(',',@counts),"]\n";
    if ($dir8==0) {
      MyFLAT::FLAT_show_breadth($f, 5, 'lowtohigh');
    }
  }

  @net_turn = map {$_/4} @net_turn;
  print "net [",join(',',@net_turn),"]\n";
  exit 0;
}





{
  my @minus_rotate_I_table = ([ 1, 4 ],  # 1 = 0
                              [ 7, 4 ],  # 2 =  1  flip
                              [ 5, 1 ],  # 3 =  I  flip
                              [ 1, 3 ],  # 4 = -1  flip
                              [ 4, 6 ],  # 5 = -I  flip
                              [ 2, 1 ],  # 6 = -1 + I
                              [ 3, 6 ]); # 7 = -1 - I
  foreach (@minus_rotate_I_table) { foreach (@$_) { $_--; } }
  require MyFLAT;
  my $f = MyFLAT::aref_to_FLAT_DFA(\@minus_rotate_I_table,
                                   accepting => 0,
                                  );
  $f->MyFLAT::view;
  exit 0;
}
{
  # ComplexMinus colours by n gray scale

  my $level = 16;
  my $width = 1400;
  my $height = 880;

  require Math::PlanePath::ComplexPlus;
  require Math::PlanePath::ComplexMinus;
  require Math::PlanePath::DragonMidpoint;
  my $path;
  $path = Math::PlanePath::ComplexPlus->new;
  $path = Math::PlanePath::ComplexMinus->new;
  $path = Math::PlanePath::DragonMidpoint->new;
  my ($n_lo, $n_hi) = $path->level_to_n_range($level);

  my $x_min = 0;
  my $x_max = 0;
  my $y_min = 0;
  my $y_max = 0;
  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    $x_min = min($x_min, $x); $x_max = max($x_max, $x);
    $y_min = min($y_min, $y); $y_max = max($y_max, $y);
  }

  my $scale = min(10,
                  int($width / ($x_max - $x_min + 1)),
                  int($height / ($y_max - $y_min + 1)));

  require GD;
  require Image::Base::GD;
  my $gd = GD::Image->new ($width,$height, 1); # truecolor 24-bit
  my $image = Image::Base::GD->new (-gd => $gd);
  $image->rectangle(0,0, $width-1,$height-1, 'black');
  my $transform = sub {
    my ($x,$y) = @_;
    $x -= $x_min;
    $y -= $y_max;
    $x++; $y--;
    $y = -$y;
    $x *= $scale;
    $y *= $scale;
    return ($x,$y);
  };
  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    ($x,$y) = $transform->($x,$y);
    my $colour = $n/$n_hi * 200 + 40;
    $colour = sprintf '#%02X%02X%02X', $colour, $colour, $colour;
    $image->rectangle($x,$y, $x+$scale-1,$y+$scale-1, $colour, 1);
  }
  {
    my ($x,$y) = $path->n_to_xy(0);
    ($x,$y) = $transform->($x,$y);
    $image->rectangle($x,$y, $x+$scale-1,$y+$scale-1, 'red', 1);
  }
  my $filename = "/tmp/minus-$level.png";
  $image->save($filename);

  require IPC::Run;
  push @ipc, IPC::Run::start(['xzgv','-f',$filename],'&');
  exit 0;
}
{
  # ComplexMinus angle range of boundary touching

  #  *<-----E
  #  ^      |*
  #  |      v
  #  *----->+<-----*
  #      *  ^      |
  #         |   *  v
  #         O----->*

  # k=15  0.445731514060755 to -0.274923962819157  delta 0.279344523120089  n=51494 to 60712
  # MinusPoint(51494) == -77 + 27*I
  # MinusPoint(60712) == -10 - 66*I
  # JEQ(16) == (153+1/2) + (51+1/2)*I
  # JEC(16) == 153 + 51*I
  # (z1 = JEC(15)*I) == 52 + 102*I
  # (z2 = 2*b^15*I - JEC(15)*I - (1+I)) == 203 + 153*I
  # DragonToPlus(52 + 102*I) == 25 - 77*I
  #
  # PlusOffsetH(15) == -51 - 102*I
  # PlusOffsetV(15) == -77 - 25*I
  # (ph = PlusToDragon(PlusOffsetH(15))) == 153 + 51*I
  # (pv = PlusToDragon(PlusOffsetV(15))) == 102 - 52*I
  # DragonToPlus(JEC(16)) == -51 - 102*I
  # DragonToPlus(JEC(16)) == -51 - 102*I
  # (z1-ph)
  # (z2-ph)
  # arg((z1-ph)/(z2-ph)) /Pi*180
  # arg((z1-pv)/(z2-pv)) /Pi*180
  # o=3+I
  # o=2+4*I
  # j1=-1+3*I
  # j2=1+(5+2)*I
  # arg((j1-o)/(j2-o)) /Pi*180


  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  foreach my $k (1 .. 24) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my ($n_next_lo, $n_next_hi) = $path->level_to_n_range($k+1);

    my (@cx, @cy, @cn2);
    my $accumulate = sub {
      my ($n2, $x,$y) = @_;
      push @cx, $x;
      push @cy, $y;
      push @cn2, $n2;
    };
    foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);

      foreach my $dir4 (0 .. 3) {
        my $dx = $dir4_to_dx[$dir4];
        my $dy = $dir4_to_dy[$dir4];

        my $n2 = $path->xy_to_n($x+1,$y);
        if ($n2 > $n_hi && $n2 <= $n_next_hi) {
          $accumulate->($n2, $x+.5, $y-.5);
          $accumulate->($n2, $x+.5, $y+.5);
        }
      }
    }

    my @angles = map { atan2($cy[$_],$cx[$_]) * 1/( 2*pi() ) } 0 .. $#cx;
    my @angles_i = sort {$angles[$a]<=>$angles[$b]} 0 .. $#angles;
    my $start_angle = 0;
    my $start_i = 0;
    my $end_angle = 0;
    my $end_i = 0;
    foreach my $pos (0 .. $#angles_i) {
      my $i = $angles_i[$pos];
      my $im = $angles_i[$pos-1];
      my $ai = $angles[$i];
      my $aim = $angles[$im];
      my $delta = $ai - $aim;
      if ($delta < 0) { $delta += 1; }
      if ($delta > .25) {
        $start_angle = $ai;
        $end_angle = $aim;
        $start_i = $i;
        $end_i = $im;
        # print "cross $delta\n";
      }
    }
    my $delta = $end_angle - $start_angle;
    my $start_n = $cn2[$start_i];
    my $end_n = $cn2[$end_i];
    if ($delta < 0) { $delta += 1; }
    print "k=$k  $start_angle to $end_angle  delta $delta  n=$start_n to $end_n\n";
  }
  exit 0;
}
{
  # complex i-1 blocks on boundary -- starting from single
  #
  #
  #  g  |  a  | f              e   |  e  |  d  |  d  |
  #     |     |                          |           |
  # ----+-----+----      ----------+-----+-----+-----+
  #     |     |                    |           |
  #  b  |  0  | e        f     f   |  0  |  1  |  c  |  c
  #     |     |                    |           |
  # ----+-----+----       ---------+-----+-----+-----
  #     |     |                          |
  #  c  |  d  | h              a   |  a  |  b  |  b
  #
  # I*(I-1) == -1-I  \\ a
  # -1*(I-1) == 1-I  \\ b
  # (-1-I)*(I-1) == 2  \\ c

  my $a = 1 << 0;
  my $b = 1 << 1;
  my $c = 1 << 2;
  my $d = 1 << 3;
  my $e = 1 << 4;
  my $f = 1 << 5;
  my $g = 1 << 6;
  my $h = 1 << 7;

  my $all = $a|$b|$c|$d|$e|$f|$g|$h;
  my $initial = 0;

  my $expand = sub {
    my ($s) = @_;
    return ((  ($s & $e) && $a)
            | (($s & $f) && $b)
            | (($s & $a) && $c)
            | (($s & $a) && $d)
            | ((1      ) && $e)
            | (($s & $d) && $f)
            | (($s & $e) && $g)
            | (($s & $b) && $h),

            (  ($s & $d) && $a)
            | ((1      ) && $b)
            | (($s & $a) && $c)
            | (($s & $b) && $d)
            | (($s & $c) && $e)
            | (($s & $d) && $f)
            | (($s & $e) && $g)
            | (($s & $b) && $h));
  };

  my $n_to_state_by_high_to_low = sub {
    my ($n, $k) = @_;
    my $s = $initial;
    my @bits = bit_split_lowtohigh($n);
    for (my $i = $k-1; $i >= 0; $i--) {
      my $bit = $bits[$i];   # high to low
      my @e = $expand->($s);
      $s = $e[$bit||0];
    }
    return $s;
  };
  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  my @path_offsets = ([ 0, 1, $a],
                      [-1, 0, $b],
                      [-1,-1, $c],
                      [ 0,-1, $d],
                      [ 1, 0, $e],
                      [ 1, 1, $f],
                      [-1, 1, $g],
                      [ 1,-1, $h]);
  my $n_to_state_by_path = sub {
    my ($n, $k) = @_;
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my ($x,$y) = $path->n_to_xy($n);
    my $s = 0;
    foreach my $elem (@path_offsets) {
      my ($dx,$dy, $s_bit) = @$elem;
      my $n_neighbour = $path->xy_to_n($x+$dx,$y+$dy);
      if (defined $n_neighbour && $n_neighbour <= $n_hi) {
        $s |= $s_bit;
      }
    }
    return $s;
  };
  foreach my $k (0 .. 10) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi) {
      my $s_expand = $n_to_state_by_high_to_low->($n, $k);
      my $s_path   = $n_to_state_by_path->($n, $k);
      # printf "k=%d n=%*b  exp=%6b  path=%6b\n", $k, $k, $n, $s_expand, $s_path;
      if ($s_path != $s_expand) {
        printf "path=%b expand=%b\n", $s_path, $s_expand;
        die;
      }
    }
  }

  my $s_to_name = sub {
    my ($s) = @_;
    if ($s == 0) { return 'zero'; }
    return join('',
                ($s & $a ? 'a' : ()),
                ($s & $b ? 'b' : ()),
                ($s & $c ? 'c' : ()),
                ($s & $d ? 'd' : ()),
                ($s & $e ? 'e' : ()),
                ($s & $f ? 'f' : ()),
                ($s & $g ? 'g' : ()),
                ($s & $h ? 'h' : ()));
  };

  {
    my $s = $all;
    my ($s0,$s1) = $expand->($s);
    printf "%b -> %b, %b\n", $s, $s0, $s1;
  }

  require MyFSM;
  my %table;
  foreach my $s (0 .. $all) {
    my ($s0,$s1) = $expand->($s);
    $table{$s_to_name->($s)}{0} = $s_to_name->($s0);
    $table{$s_to_name->($s)}{1} = $s_to_name->($s1);
  }

  my %accepting;
  my $title;

  if (1) {
    $accepting{$s_to_name->($all)} = 'non';
    $title = 'accept all';
  }
  if (0) {
    # accepting a,b,d,e
    foreach my $s (0 .. $all) {
      if (($s & $a)
          && ($s & $b)
          && ($s & $d)
          && ($s & $e)) {
        $accepting{$s_to_name->($s)} = 'non';
      }
    }
    $title = 'i-1 NSEW';
  }
  if (0) {
    # accepting any "a" side non-boundary
    my $target = $a;
    foreach my $s (0 .. $all) {
      if (($s & $target) != 0) {
        $accepting{$s_to_name->($s)} = 'non';
      }
    }
    $title = 'accept '.$s_to_name->($target);
  }

  my $fsm = MyFSM->new(table => \%table,
                       initial => $s_to_name->($initial),
                       accepting => \%accepting,
                       title => $title);

  # $fsm->merge_state('abcdef-prev0','abcdef-prev1');
  # $fsm->rename_state('abcdef-prev0','abcdef');
  $fsm = $fsm->reverse;
  $fsm->simplify;
  $fsm->renumber_states;
  # $fsm = $fsm->digit_delta;

   $fsm->rename_state(16,'a');
   $fsm->rename_state(14,'d');
   $fsm->rename_state(20,'b');
   $fsm->rename_state(19,'e');
   $fsm->rename_state(22,'f');
   $fsm->rename_state(21,'c');
   $fsm->rename_state(13,'s14');
   $fsm->rename_state(18,'s7');
   $fsm->rename_state(11,'s13');
   $fsm->rename_state(8,'s8');

  # $fsm->delete_unreached_states;
  $fsm->print_tikz;
  $fsm->show;
  $fsm->view;
  exit;
  {
    foreach my $k (0 .. 6) {
      print "k=$k\n";
      foreach my $n (0 .. 2**$k-1) {
        if (complex_minus_n_is_six($n,$k)) {
          my $delta_n = delta_n($n,$k);
          printf "%*b %*b\n", $k,$n, $k,$delta_n;
        }
      }
    }
  }
  {
    print $fsm->{'title'},"\n";
    foreach my $k (0 .. 6) {
      print "k=$k\n";
      my @sequence_list = $fsm->accepting_sequences($k);
      @sequence_list = map{join(',',@$_)} @sequence_list;
      @sequence_list = sort @sequence_list;
      foreach my $sequence (@sequence_list) {
        print $sequence,"\n";
      }
    }
  }

  $fsm = $fsm->reverse;
  # $fsm->delete_unreached_states;
  $fsm->simplify (verbose => 0);
  # $fsm->simplify;
  $fsm->renumber_states;

  # $fsm = $fsm->digit_delta;
  # $fsm->merge_state('non-prev0','non-prev1');

  # $fsm->rename_state(5,'z1');
  # $fsm->rename_state(1,'z2');
  # $fsm->rename_state(7,'z3');
  # $fsm->rename_state(8,'z4');
  # $fsm->rename_state(2,'z5');
  # $fsm->rename_state(3,'z8');
  # $fsm->rename_state(6,'z6');

  # {
  #   print $fsm->{'title'},"\n";
  #   foreach my $k (0 .. 6) {
  #     print "k=$k\n";
  #     my @sequence_list = $fsm->accepting_sequences($k);
  #     @sequence_list = map{ [ reverse @$_ ] } @sequence_list;
  #     @sequence_list = map{join(',',@$_)} @sequence_list;
  #     @sequence_list = sort @sequence_list;
  #     foreach my $sequence (@sequence_list) {
  #       print $sequence,"\n";
  #     }
  #   }
  # }

  $fsm->view;
  $fsm->show;
  $fsm->print_tikz;

  exit 0;
}
{
  # complex i-1 NSEW by low to high

  # to 5  00
  #       10
  # to 4  00
  my @table = ([1,2],  # 0
               [5,3],
               [1,5],  # 2
               [4,2],
               [5,6],  # 4
               [6,5],
               [8,7],  # 6
               [4,5],
               [9,9],  # 8
               [9,9]); # 9 non
  my $nsew_by_low_to_high_state = sub {
    my ($n, $k) = @_;
    {
      my @bits = delta_bit_split_lowtohigh($n,$k);
      ### @bits
      my $state = 0;
      foreach my $bit (@bits) {
        $state = $table[$state][$bit];
        ### $state
      }
      return $state;
    }
  };
  my $nsew_by_low_to_high = sub {
    my ($n, $k) = @_;
    return $nsew_by_low_to_high_state->($n,$k) == 9;
  };

  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;

  foreach my $k (0 .. 18) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi) {
      my $nsew_expand = $nsew_by_low_to_high->($n, $k);
      my $nsew_path   = complex_minus_n_is_nsew($n, $k);
      my $state = $nsew_by_low_to_high_state->($n,$k);
      # printf "k=%d n=%*b dstr=%*b  exp=%d[%d] path=%d\n",
      #   $k, $k,$n, $k,delta_n($n,$k), $nsew_expand,$state, $nsew_path;
      if ($nsew_path != $nsew_expand) {
        die;
      }
    }
  }
  exit 0;
}
{
  # complex i-1 6 neighbours by low to high
  my $deltastr;
  my $six_by_low_to_high = sub {
    my ($n, $k) = @_;


    my @bits = bit_split_lowtohigh($n);
    while (@bits < $k) { push @bits, 0; } # pad
    { # delta
      my $prev = 0;
      foreach my $i (reverse 0 .. $#bits) {
        my $bit = $bits[$i];
        $bits[$i] = $bit ^ $prev;  # if different
        $prev = $bit;
      }
    }
    $deltastr = join(' ',map {$_?1:0} @bits);
    ### $deltastr

    {
      foreach (1 .. 4) {
        if (shift @bits) { return 0; }
      }
      return @bits > 0;
    }
    {
      my $state;
      if (shift @bits)  {
        while (shift @bits) { shift @bits; }
        $state = 2;
      } else {
        $state = 6;
      }
      ### initial: $state
      while (@bits) {
        my $bit = shift @bits;
        if ($state == 2) {
          $state = (!$bit ? 6 : 3);
        } elsif ($state == 6) {
          $state = (!$bit ? 3 : 2);
        } elsif ($state == 3) {
          $state = (!$bit ? 4 : 2);
        } elsif ($state == 4) {
          $state = (!$bit ? 1 : 2);
        } elsif ($state == 1) {
          return 1;
        } else {
          die $state;
        }
        ### $state
      }
      # if ($state == 1) { return 1; }
      ### not surrounded ...
      return 0;
    }
  };

  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;

  foreach my $k (0 .. 10) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi) {
      my $six_expand = $six_by_low_to_high->($n, $k);
      my $six_path   = complex_minus_n_is_six($n, $k);
      printf "k=%d n=%*b dstr=%s  exp=%d path=%d\n",
        $k, $k,$n, $deltastr, $six_expand, $six_path;
      if ($six_path != $six_expand) {
        die;
      }
    }
  }
  exit 0;
}


{
  # complex i+1 blocks on boundary
  #
  #  +--------+
  #  |   e    |
  #  |f zero g|
  #  |   a    |
  #  +--------+
  #
  #  +-------+-------+
  #  |   e   |   d   |
  #  |f Even | Odd  c|
  #  |   a   |   b   |
  #  +-------+-------+
  #

  my $a = 1 << 0;
  my $b = 1 << 1;
  my $c = 1 << 2;
  my $d = 1 << 3;
  my $e = 1 << 4;
  my $f = 1 << 5;

  my $rotate = sub {
    my ($s) = @_;
    return ((  ($s & $a) && $d)
            | (($s & $b) && $e)
            | (($s & $c) && $f)
            | (($s & $d) && $a)
            | (($s & $e) && $b)
            | (($s & $f) && $c));
  };

  my $expand = sub {
    my ($s) = @_;
    return ((  ($s & $f) && $a)
            | (($s & $a) && $b)
            | (($s & $b) && $c)
            | (($s & 0 ) && $d)
            | (($s & $e) && $e)
            | (($s & $e) && $f),

            (  ($s & 0 ) && $a)
            | (($s & $b) && $b)
            | (($s & $b) && $c)
            | (($s & $c) && $d)
            | (($s & $d) && $e)
            | (($s & $e) && $f));
  };
  my $s_to_name = sub {
    my ($s) = @_;
    if ($s == 0) { return 'non'; }
    return join('',
                ($s & $a ? 'a' : ()),
                ($s & $b ? 'b' : ()),
                ($s & $c ? 'c' : ()),
                ($s & $d ? 'd' : ()),
                ($s & $e ? 'e' : ()),
                ($s & $f ? 'f' : ()));
  };

  my $all = $a|$b|$c|$d|$e|$f;
  my $initial = $all;

  {
    my $s = $all;
    my ($s0,$s1) = $expand->($s);
    printf "%b -> %b, %b\n", $s, $s0, $s1;
  }
  {
    my $s = $a+$b+$c+$e+$f;
    my $s2 = $rotate->($s);
    printf "rotate %s -> %s\n", $s_to_name->($s), $s_to_name->($s2);
  }

  require MyFSM;
  my %table;
  foreach my $s (0 .. $all) {
    my ($s0,$s1) = $expand->($s);
    $table{$s_to_name->($s)}{0} = $s_to_name->($s0);
    $table{$s_to_name->($s)}{1} = $s_to_name->($s1);
  }

  my %accepting;
  $accepting{$s_to_name->(0)} = 'non';

  if (0) {
    # accepting any "c" side non-boundary
    foreach my $s (0 .. $all) {
      if (($s & $c) == 0) {
        $accepting{$s_to_name->($s)} = 'non';
      }
    }
  }
  if (0) {
    # accepting any "b" side non-boundary
    foreach my $s (0 .. $all) {
      if (($s & $b) == 0) {
        $accepting{$s_to_name->($s)} = 'non';
      }
    }
  }
  if (0) {
    # accepting any "a" side non-boundary
    foreach my $s (0 .. $all) {
      if (($s & $a) == 0) {
        $accepting{$s_to_name->($s)} = 'non';
      }
    }
  }

  my $fsm = MyFSM->new(table => \%table,
                       initial => $s_to_name->($initial),
                       accepting => \%accepting);

  # {
  #   my $count = 0;
  #   my @done;
  #   foreach my $s (0 .. $all) {
  #     my $s2 = $rotate->($s);
  #     next if $done[$s]++;
  #     next if $done[$s2]++;
  #     if ($s != $s2) {
  #       print "merge: $s $s2\n";
  #       $fsm->merge_state($s_to_name->($s), $s_to_name->($s2));
  #       $fsm->rename_state($s_to_name->($s),
  #                          $s_to_name->($s).'+'.$s_to_name->($s2));
  #       # last if $count++ > 6;
  #     }
  #   }
  # }

  $fsm->simplify;
  # $fsm->delete_unreached_states;
  print $fsm->as_graph6 (header => 0);
  print $fsm->as_sparse6 (header => 0);
  $fsm->print_tikz;
  $fsm->show;
  $fsm->view;

  $fsm = $fsm->reverse;
  $fsm->simplify;
  my $n = 1;
  foreach my $s ($fsm->sorted_states) {
    print "$s\n";
    if ($s eq $fsm->{'initial'}) {
      $fsm->rename_state($s, 'initial');
    } elsif ($s =~ /^(non,?)*$/) {
      $fsm->rename_state($s, 'non');
    } else {
      $fsm->rename_state($s, $n++);
    }
  }
  $fsm->view;
  $fsm->print_tikz;
  print $fsm->as_graph6 (header => 0);
  print $fsm->as_sparse6 (header => 0);

  exit 0;
}
{
  # Complex base i-1 X axis contiguous width

  # 0 1 2 3 4 5 6 7 8
  # 1,1,2,4,6,2,2,6,2,2,2,2,2,2,2,2,2,2,2,2,2
  #
  #                             2 3             2 3     Y=1
  # k=0  0     k=1  0 1    k=2    0 1     k=3     0 1
  #                                             6 7
  #                                               4 5   Y=-2
  #
  # k=4       10 11       Y=3      k=5         26 27      10 11       Y=3
  #               8  9                            24 25       8  9
  #       2 3 14 15       Y=1            18 19 30 31  2 3 14 15       Y=1
  #         0  1 12 13                      16 17 28 29 0  1 12 13
  #       6 7             Y=-1           22 23        6 7             Y=-1
  #         4  5          Y=-2              20 21       4  5          Y=-2
  #

  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  my @values;
  foreach my $k (0 .. 20) {
    print "k=$k\n";
    (my $n_lo, my $n_hi) = $path->level_to_n_range($k);
    my $limit = 5*2**(int(($k+2)/2));

    my $x_max = 0;
    my $x_min = 0;
    foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      $x_min = min($x_min, $x);
      $x_max = max($x_max, $x);
    }

    my $why;
    my $y_is_contiguous = sub {
      my ($y) = @_;
      my $runs = 0;
      my $prev = 0;
      $why = '';
      foreach my $x ($x_min-2 .. $x_max+2) {
        my $n = $path->xy_to_n($x,$y);
        my $this = (defined $n && $n <= $n_hi);
        if ($this && ! $prev) {
          $runs++;
          if ($runs >= 2) {
            $why = "second run starting $x";
            return 0;
          }
        }
        $prev = $this;
      }
      if ($runs != 1) {
        $why = "no runs at all";
        return 0;
      }
      return 1;
    };

    my $width = 0;
    for (my $y = 0; ; $y++) {
      if (! $y_is_contiguous->($y)) {
        print "  y=$y $why\n";
        last;
      }
      $width++;
    }
    for (my $y = -1; ; $y--) {
      if (! $y_is_contiguous->($y)) {
        print "  y=$y $why\n";
        last;
      }
      $width++;
    }

    print "  width $width      X=$x_min to $x_max\n";
    push @values, $width;
  }

  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values,
                           verbose => 1);
  exit 0;
}
{
  # Complex base axes continuous
  # ComplexMinus continuous axes and diagonals
  #
  # ComplexPlus runs
  # 1,1,1,2,3,3,3,6,9,9,9,18,27,27,27,54,81,81,81
  # unduplicated A038754 a(2n) = 3^n, a(2n+1) = 2*3^n.
  #
  #  0           k=0  1 each
  #  0  1        k=1  1 each
  #
  #     2  3     k=2  1 each
  #  0  1
  #
  #     6  7
  #  4  5
  #     2  3     k=3  2 Y, 1 other
  #  0  1
  #
  #         14 15
  #      12 13
  #         10 11  6  7
  #       8  9  4  5
  #                2  3     k=3  Y=3, opp=2
  #             0  1
  #
  # DragonCurve runs, unduplicated
  # X axis   1,2,5,14,41,122,365 = (3^n + 1)/2
  # Y axis   1,3,8,9,23,27,68,81,203,243      3^n and 3^n - (3^(n-X) - 1)/2
  # leading  1,2,5,14,41,122,365 = X axis
  # opposite 1,3,8,9,23,27,68,81,203,243 = Y axis
  #
  # b^k      1,2,3,5,8,14,23,41,68,122,203,365
  #             = A246360 (3^n+1)/2 and (5*3^(n-1)+1)/2
  #          1,2,3,5,9,14,27,41,81,122,243
  #          1,2,3,5,9,14,27,41,81,122,243
  #          1,2,3,5,9,14,27,41,81,122,243,365
  require Math::PlanePath::ComplexPlus;
  my $path = Math::PlanePath::ComplexPlus->new;
   $path = Math::PlanePath::DragonCurve->new;
  my @values;
  foreach my $k (0 .. 23) {
    (my $n_lo, my $n_hi) = $path->level_to_n_range($k);
    my $limit = 2**(int(($k+2)/2));
    my $count_runs = sub {
      my ($func) = @_;
      my $prev = 0;
      my $runs = 0;
      foreach my $i (-$limit .. $limit) {
        my ($x,$y) = $func->($i);
        my $n = $path->xy_to_n($x,$y);
        my $this = (defined $n && $n <= $n_hi);
        if ($prev && ! $this) {
          $runs++;
        }
        $prev = $this;
      }
      return $runs;
    };

    my @count = ($count_runs->(sub{return ($_[0], 0)}),  # X
                 $count_runs->(sub{return ($_[0], $_[0])}),  # leading
                 $count_runs->(sub{return (0, $_[0])}),  # Y
                 $count_runs->(sub{return ($_[0], -$_[0])}),  # opposite
                );

    # relative to direction of b^k
    foreach (1 .. $k) {
      push @count, shift @count;
    }

    print "k=$k  ",join(' ',map{sprintf'%4d',$_}@count),"\n";

    foreach my $i (0 .. $#count) {
      push @{$values[$i]}, $count[$i];
    }
  }

  require Math::OEIS::Grep;
  require List::MoreUtils;
  foreach my $i (0 .. $#values) {
    my $aref = $values[$i];
    @$aref = List::MoreUtils::uniq(@$aref);
    Math::OEIS::Grep->search(array => $aref,
                             name => "i=$i",
                             verbose => 1);
  }
  exit 0;
}
{
  # MinusNeg etc pieces by turns Four or Eight

  # recurrence_guess()

  # k=8 turn +1   12 of
  #
  # MinusUnpoint(1) == 1
  # MinusUnpoint(2) == 12
  # MinusDoublePred(8,1)
  # MinusDoublePred(8,12)
  #
  # MinusUnpoint(3+3*I) == 102
  # MinusUnpoint(3+3*I+1) == 103
  # MinusUnpoint(3+3*I+I) == 125
  # MinusUnpoint(3+3*I+1+I) == 96
  # MinusInc(102,1) == 103
  # MinusInc(102,I) == 125
  # MinusInc(102,1+I) == 96
  # MinusDoublePred(8,102) == 1
  # MinusDoublePred(8,103) == 0
  # MinusDoublePred(8,125) == 0
  # MinusDoublePred(8,96)  == 0
  # recurrence_guess()

  my $max_k = to_bigint(90);
  my @net_turn = (0) x $max_k;

  my $type = 'Complex';
  $type = 'MinusNeg';
  $type = 'MinusRot';
  $type = 'MinusDouble';
  my $eight = 0;

  my $pred = Pred($type);
  # none a,b,diagonal,  turn +1
  foreach my $dir8 (0,4,2,6) {
    my $dir8_plus1 = ($dir8+1) % 8;
    my $dir8_plus2 = ($dir8+2) % 8;
    my $f = $pred
      ->intersect(NeighbourPred($type,$dir8)->complement)
      ->intersect(NeighbourPred($type,$dir8_plus2)->complement);
    # Four diagonal don't care, Eight must be absent
    if ($eight) {
      $f = $f->intersect(NeighbourPred($type,$dir8_plus1)->complement);
    }
    print "turn+1 states ",$f->num_states,"\n";
    my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $net_turn[$i] += $counts[$i]; }
    print "dir8=$dir8 turn+1 [",join(',',@counts),"]\n";
    # if ($dir8==0) {
    #   MyFLAT::FLAT_show_breadth($f, 8, 'lowtohigh');
    # }
  }

  if ($eight) {
    # with NE only, turn -1
    foreach my $dir8 (0,4,2,6) {
      my $dir8_plus1 = ($dir8+1) % 8;
      my $dir8_plus2 = ($dir8+2) % 8;
      my $f = $pred
        ->intersect(NeighbourPred($type,$dir8)->complement)
        ->intersect(NeighbourPred($type,$dir8_plus1))
        ->intersect(NeighbourPred($type,$dir8_plus2)->complement);
      print "turn-1 states ",$f->num_states,"\n";
      my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
      foreach my $i (0 .. $max_k) { $net_turn[$i] -= $counts[$i]; }
      print "dir8=$dir8 2*turn-1 [",join(',',@counts),"]\n";
    }
  }

  # both a,b  and no diagonal turn -1
  foreach my $dir8 (0,4,2,6) {
    my $dir8_plus1 = ($dir8+1) % 8;
    my $dir8_plus2 = ($dir8+2) % 8;
    my $f = $pred
      ->intersect(NeighbourPred($type,$dir8))
      ->intersect(NeighbourPred($type,$dir8_plus1)->complement)
      ->intersect(NeighbourPred($type,$dir8_plus2));
    print "both states ",$f->num_states,"\n";
    my @counts = MyFLAT::FLAT_count_contains($f,$max_k);
    foreach my $i (0 .. $max_k) { $net_turn[$i] -= $counts[$i]; }
    print "dir8=$dir8 both   [",join(',',@counts),"]\n";
    # if ($dir8==0) {
    #   MyFLAT::FLAT_show_breadth($f, 5, 'lowtohigh');
    # }
  }

  @net_turn = map {$_/4} @net_turn;
  print "net [",join(',',@net_turn),"]\n";
  exit 0;
}
{
  # MinusNeg number of pieces Four by path
  #
  # recurrence_guess([0,1,1,1,1,4,7,7,8,13,29,50,61])
  # not in OEIS: 1,4,7,7,8,13,29,50,61

  my $path = Math::PlanePath::ComplexMinus->new;
  my $prev = 0;
  my @values;
 K: foreach my $k (0 .. 12) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my @n_to_piece;
    my $upto_piece = 0;
    my %pieces;
  N: foreach my $n (0 .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      next unless $path->xy_to_n(-$x,-$y) <= $n_hi;
      my $piece = $upto_piece++;
      $n_to_piece[$n] = $piece;
      ### at: "n=$n  xy=$x,$y  piece=$piece"

      $pieces{$piece} = 1;
      foreach my $dir (0 .. 3) {
        my $dx = $dir4_to_dx[$dir];
        my $dy = $dir4_to_dy[$dir];
        my $x2 = $x+$dx;
        my $y2 = $y+$dy;
        my $n2 = $path->xy_to_n($x2,$y2);
        next unless $n2 <= $n_hi;
        next unless $path->xy_to_n(-$x2,-$y2) <= $n_hi;
        if (defined(my $p = $n_to_piece[$n2])) {
          ### unify: $p
          delete $pieces{$p};
          foreach (@n_to_piece) {
            if (defined $_ && $_ == $p) { $_ = $piece; }  # mutate array
          }
        }
      }
    }
    my $num_pieces = scalar keys %pieces;
    my $diff = $num_pieces - $prev;
    print "k=$k  $num_pieces  +$diff\n";
    $prev = $num_pieces;
    push @values, $num_pieces;
  }
  print "pieces [",join(',',@values),"]\n";
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}

BEGIN {
  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  my @dir_to_dx = (1, 1, 0,   -1, -1,  0 );
  my @dir_to_dy = (0, 1, 1,    0, -1, -1 );
  sub complex_minus_n_is_six {
    my ($n, $k) = @_;
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    ### $n_hi
    my ($x,$y) = $path->n_to_xy($n);
    foreach my $dir (0 .. $#dir_to_dx) {
      my $n_neighbour = $path->xy_to_n($x + $dir_to_dx[$dir],
                                       $y + $dir_to_dy[$dir]);
      unless (defined $n_neighbour && $n_neighbour <= $n_hi) {
        return 0;
      }
    }
    return 1;
  };
}

BEGIN {
  require Math::PlanePath::ComplexMinus;
  my $path = Math::PlanePath::ComplexMinus->new;
  sub complex_minus_n_is_nsew {
    my ($n, $k) = @_;
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    ### $n_hi
    my ($x,$y) = $path->n_to_xy($n);
    foreach my $dir4 (0 .. $#dir4_to_dx) {
      my $n_neighbour = $path->xy_to_n($x + $dir4_to_dx[$dir4],
                                       $y + $dir4_to_dy[$dir4]);
      unless (defined $n_neighbour && $n_neighbour <= $n_hi) {
        return 0;
      }
    }
    return 1;
  };
}

