#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use Image::Base::GD;
use List::Util 'min','max';
use Math::PlanePath::DragonCurve 118; # v.118 for xyxy_to_n_either()

use FindBin;
use lib "$FindBin::Bin/../tools";
use MyDragon 'xy_rotate_minus90','transform_to_horizontal';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 16;
if (@ARGV) { $level = shift @ARGV; }

($level & 1) == 0 or die "level must be even";


sub weight_to_colour {
  my ($t) = @_;

  # range 0 to 255 linear
  {
    $t /= 1200;  # ranging 0 to 1 inclusive
    $t = int(255*$t);
    $t = 255 - $t;  # black curve, white background
    return sprintf '#%02X%02X%02X', $t, $t, $t;
  }

  # black or white
  {
    $t /= 1200;  # ranging 0 to 1 inclusive
    return $t >= .5 ? 'black' : 'white';
  }

  # range 10 to 100 parabolic
  {
    $t /= 1200;  # ranging 0 to 1 inclusive
    $t = ($t+.5)**2;  # .5**2 = .25 to 1.5**2 = 2.25
    $t = ($t - .25) / 2;  # 0 to 1
    my $percent = int(10 + 90*$t);
    return "[black!$percent]";
  }

  # range 10 to 100 linear
  {
    $t /= 1200;  # ranging 0 to 1 inclusive
    my $percent = int(10 + 90*$t);
    return "[black!$percent]";
  }
}

my $path = Math::PlanePath::DragonCurve->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);

my $wx_min = 0;
my $wy_min = 0;
my $wx_max = 0;
my $wy_max = 0;
my %weight;

sub add {
  my ($wx,$wy, $weight) = @_;
  $weight{$wx}{$wy} += $weight;
  $wx_min = min($wx_min, $wx);
  $wy_min = min($wy_min, $wy);
  $wx_max = max($wx_max, $wx);
  $wy_max = max($wy_max, $wy);
}

sub transform {
  my ($x,$y) = @_;
  ($x,$y) = transform_to_horizontal($x,$y, $level);
  return ($x,$y);
}

foreach my $n ($n_lo .. $n_hi-1) {
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  if ($n & 1) {
    ($x,$y) = ($x+$dx,$y+$dy);
    ($dx,$dy) = (-$dx,-$dy);
  }
  my ($rx,$ry) = xy_rotate_minus90($dx,$dy);

  add (2*$x -   $dx + $rx, 2*$y -   $dy + $ry,  96);
  add (2*$x +   $dx + $rx, 2*$y +   $dy + $ry,  348);
  add (2*$x + 3*$dx + $rx, 2*$y + 3*$dy + $ry,  6);

  add (2*$x -   $dx - $rx, 2*$y -   $dy - $ry,  24);
  add (2*$x +   $dx - $rx, 2*$y +   $dy - $ry,  102);
  add (2*$x + 3*$dx - $rx, 2*$y + 3*$dy - $ry,  24);
}


# 
# GP-Test  1/3+1+1/6 == 3/2   /* horizontal */
#
my $len = 2**($level/2); #  start to end
my $width = int(3/2 * $len);
my $height = int($len);
my $border = 10;
$width += 2*$border;
$height += 2*$border;
my $size = 16;
my $x_base = int(-1/2 * $len) - $border;
my $y_base = int(-2/3 * $len) - $border;

### $x_base
### $y_base

my $image = Image::Base::GD->new (-height => $height,
                                  -width  => $width);
$image->rectangle(0,0, $width-1,$height-1, 'white');

my %colours;
my $count_squares = 0;
foreach my $wy ($wy_min .. $wy_max) {
  foreach my $wx ($wx_min .. $wx_max) {
    my $weight = $weight{$wx}{$wy} || next;
    my $colour = weight_to_colour($weight);
    ### $colour
    $colours{$colour} = 1;

    my ($x1,$y1) = ($wx-1,$wy-1);
    my ($x2,$y2) = ($wx+1,$wy+1);
    ($x1,$y1) = transform($x1,$y1);
    ($x2,$y2) = transform($x2,$y2);
    my $mx = min($x1,$x2);
    my $my = min($y1,$y2);
    $mx /= 2;
    $my /= 2;
    $mx -= $x_base;
    $my -= $y_base;
    ### draw: "$mx, $my  $colour"
    $image->xy($mx,$height-1-$my, $colour);
    $count_squares++;
  }
}

my $num_distinct_colours = scalar keys %colours;
print "${width}x$height  $count_squares squares, $num_distinct_colours distinct colours\n";

my $filename = '/tmp/x.png';
$image->save($filename);

require IPC::Run;
IPC::Run::run(['xzgv',$filename]);

# {
#   print "  % weights: ",
#     my $join = '';
#   foreach my $weight (0 .. $#histogram) {
#     next unless $histogram[$weight];
#     my $t = $weight / 1200 * 100;
#     printf "%s%.1f", $join, $t;
#     $join = ', ';
#   }
#   print "\n";
# }

exit 0;
