/* Copyright 2021 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Usage: gcc -O9 xxx.c

   ./a.out highs                   # record high runs
   ./a.out pos                     # position those highs
   ./a.out runs     >b083279.txt   # A083279 all runs
   ./a.out parities                # A061419 mod 2 parities
   ./a.out b205083  >b205083.txt   # A205083 parities with extra initial 1

   In the code, OUTPUT_LIMIT is the last index to print in the output.
   This caps the "runs" and "parities" output types.  The record highs are
   too few to reach this limit.

   Values of w = A061419 are calculated as bignums in an array of words.
   Parities are w mod 2 = A205083 (except only one initial 1, unlike two
   there).  Run lengths = A083279 are runs of 1s or of 0s in the parities.

   Values w are roughly (3/2)^n so each megabyte of memory for w is about n
   = 13 000 000 parities terms, then various run lengths occur.

   The calculation ceil(w*3/2) is a simple right shift down and add,
   ie. w + w/2 rounded up.  This takes quadratic time in the length of w.

   The full value of w is needed, in general, since successive w/2 means
   that every bit of a given w works its way down to affect later w mod 2.
   (Knowing a bit of w at bit position k will take k steps until reaching
   w mod 2 would mean that if a pre-determined limit then w can be truncated
   to the parts of w low enough to influence w mod 2 up to that limit.
*/

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define OUTPUT_LIMIT  1000
#define W_SIZE  (10*1000*1000)  /* words */

typedef unsigned long  WORD;  /* an unsigned type */
#define WORD_BITLEN  (sizeof(unsigned)*CHAR_BIT)

#ifdef __GNUC__
#define LIKELY(cond)    __builtin_expect((cond) != 0, 1)
#define UNLIKELY(cond)  __builtin_expect((cond) != 0, 0)
#else
#define LIKELY(cond)    (cond)
#define UNLIKELY(cond)  (cond)
#endif

/* w[] is an array of words which is a bignum integer.
   w[0] is the least significant WORD.
   w_above is a pointer to the word immediately above the highest in w
   currently used.  (So the length of the number, but as a pointer.)

   next_w() steps w = ceil(w * 3/2) and returns 1 if fits in w[], or returns
   0 if it fits but have reached the limit of w[].  This is when w_above
   points to the last word in w[].  (This last word is yet unused, but
   stopping there keeps w_above strictly within the extent of w[].)
*/
static WORD w[W_SIZE];
WORD *w_above;

int
next_w (void)
{
  WORD t = w[0];
  WORD c = t&1;
  WORD *p = w;
  do {
    WORD next_t = p[1];
    WORD orig_t = t;
    t += c + (t>>1) + (next_t << (WORD_BITLEN-1));
    c = t < orig_t;
    *p++ = t;
    t = next_t;
  } while (LIKELY( p != w_above ));

  if (UNLIKELY( c )) {
    *w_above++ = c;
    if (UNLIKELY( w_above >= w + W_SIZE - 2 )) {
      return(0);
    }
  }
  return(1);
}

int
main (int argc, char *argv[])
{
#define OUTPUT_HIGHS      1
#define OUTPUT_HIGHPOS    2
#define OUTPUT_RUNS       3
#define OUTPUT_PARITIES   4
  int output_type = OUTPUT_HIGHS;
  unsigned output_n = 1;  /* index n in the output */

  setbuf(stdout,0);
  for (argv++; *argv!=NULL; argv++) {
    if (strcmp(*argv,"highs") == 0)         { output_type = OUTPUT_HIGHS; }
    else if (strcmp(*argv,"pos") == 0)      { output_type = OUTPUT_HIGHPOS; }
    else if (strcmp(*argv,"runs") == 0)     { output_type = OUTPUT_RUNS; }
    else if (strcmp(*argv,"parities") == 0) { output_type = OUTPUT_PARITIES; }
    else if (strcmp(*argv,"b205083") == 0) {
      /* A205083 has an extra initial 1 */
      printf ("%u %u\n", output_n++, 1);
      output_type = OUTPUT_PARITIES;
    } else {
      fprintf(stderr,"unrecognised argument: %s\n",*argv);
      exit(1);
    }
  }

  w[0] = 1;      /* initial value w=1 */
  w_above = w+1;
  {
    WORD prev_wparity = w[0] & 1;
    unsigned long run_index = 1;  /* first run index 1 */
    unsigned runlen = 1;          /* length so far of current run */
    unsigned max_runlen = 0;
    for (;;) {
      if (output_n > OUTPUT_LIMIT) break;
      if (output_type == OUTPUT_PARITIES) {
        printf ("%u %u\n", output_n++, (unsigned) w[0] & 1);
      }

      if (! next_w()) break;
      WORD wparity = w[0] & 1;

      if (wparity == prev_wparity) {
        runlen++;
      } else {
        if (output_type == OUTPUT_RUNS) {
          printf ("%u %u\n", output_n++, runlen);
        }
        if (UNLIKELY( runlen > max_runlen )) {
          max_runlen = runlen;
          if (output_type == OUTPUT_HIGHS) {
            printf ("%u %u\n", output_n++, runlen);
          } else if (output_type == OUTPUT_HIGHPOS) {
            printf ("%u %lu\n", output_n++, run_index);
          }
        }
        runlen = 1;
        run_index++;
      }
      prev_wparity = wparity;
    }
  }
  return (0);
}
