# Copyright 2019, 2020, 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Reading and writing AT&T format state machines and transducers.

package MyATT;
use 5.006;
use strict;
use warnings;
use List::Util 'max','sum';
use Carp 'croak';
use Clone 'clone';

# uncomment this to run the ### lines
# use Smart::Comments;

sub new {
  my $class = shift;
  return bless { arcs => [],
                 max_state => -1,
                 accepting => [],
                 @_ }, $class;
}
sub max_state {
  my ($self) = @_;
  return $self->{'max_state'};
}
sub num_states {
  my ($self) = @_;
  return $self->max_state + 1;
}
sub get_accepting {
  my ($self) = @_;
  return @{$self->{'accepting'}};
}
sub num_accepting {
  my ($self) = @_;
  return scalar(@{$self->{'accepting'}});
}
sub set_accepting {
  my $self = shift;
  $self->{'accepting'} = [ @_ ];
  # $self->{'max_state'} = max($self->{'max_state'}, @_);
  delete $self->{'accepting_hash'};
  return $self;
}
sub add_arc {
  my $self = shift;
  push @{$self->{'arcs'}}, [@_];
  $self->{'max_state'} = max($self->{'max_state'}, @_[0..1]);
  delete $self->{'arc_hash'};
  return $self;
}
sub read_file {
  my ($self, $filename) = @_;
  unless (ref $self) { $self = $self->new; }
  open my $fh, '<', $filename or croak "Cannot read $filename: $!";
  my $fa = $self->read_fh($fh);
  close $fh or croak "Error reading $filename: $!";
  return $fa;
}
sub read_fh {
  my ($self, $fh) = @_;
  unless (ref $self) { $self = $self->new; }
  while (defined(my $line = readline $fh)) {
    chomp $line;
    if (my ($from,$to,$symbol,$output)
        = ($line =~ /^(\d+)\s+(\d+)\s+(\S+)(?:\s+(\S+))?$/)) {
      $self->add_arc($from,$to,$symbol,$output);
    } elsif (my ($state) = $line =~ /^(\d+)$/) {
      push @{$self->{'accepting'}}, $state;
      # $self->{'max_state'} = max($self->{'max_state'}, $state);
    } else {
      croak "Unrecognised AT&T line: ",$line;
    }
  }
  return $self;
}

sub write_file {
  my ($self, $filename) = @_;
  open my $fh, '>', $filename or croak "Cannot write $filename: $!";
  $self->write_fh ($fh);
  close $fh or croak "Error writing $filename: $!";
}
sub write_fh {
  my ($self, $fh, %options) = @_;
  foreach my $arc (@{$self->{'arcs'}}) {
    print $fh join("\t", @$arc, ($#$arc < 3 ? ($arc->[-1]) : ())), "\n";
  }
  foreach my $state ($self->get_accepting) {
    print $fh $state,"\n";
  }
}
sub as_count_matrix {
  my ($self, %options) = @_;
  my $transpose = $options{'transpose'};
  my @m;
  foreach my $arc (@{$self->{'arcs'}}) {
    my ($from,$to) = @$arc;
    if ($transpose) {
      $m[$to]->[$from]++;
    } else {
      $m[$from]->[$to]++;
    }
  }
  return \@m;
}
sub as_FLAT {
  my ($self, %options) = @_;
  my $fa = FLAT::DFA->new;
  $fa->add_states($self->num_states);
  foreach my $arc (@{$self->{'arcs'}}) {
    my ($from,$to,$symbol,$output) = @$arc;
    if ($options{'transducer'}) {
      $symbol .= $output;
    }
    $fa->add_transition($from,$to,$symbol);
  }
  $fa->set_starting(0);
  $fa->set_accepting($self->get_accepting);
  return $fa;
}

sub arc_hash {
  my ($self) = @_;
  return ($self->{'arc_hash'} ||= do {
    my %hash;
    foreach my $arc (@{$self->{'arcs'}}) {
      my ($from,$to,$symbol) = @$arc;
      $hash{$from}->{$symbol} = $to;
    }
    \%hash;
  });
}

sub transitions {
  my $self = shift;
  my $state = shift;
  my $hash = $self->arc_hash;
  while (defined $state && @_) {
    $state = $hash->{$state}->{(shift)};
  }
  return $state;
}

# $att->contains(symbol,symbol,...)
# Return true if a string of the given symbols is accepted.
sub contains {
  my $self = shift;
  my $state = $self->transitions(0, @_);
  return (defined $state && $self->is_accepting($state));
}

sub accepting_hash {
  my ($self) = @_;
  return ($self->{'accepting_hash'} ||= do {
    ({ map {$_=>1} $self->get_accepting })
  });
}

sub is_accepting {
  my ($self, $state) = @_;
  return $self->accepting_hash->{$state};
}

sub count_contains {
  my ($self, $max_len, %options) = @_;
  my @accepting = $self->get_accepting;
  my $zero = $max_len*0;  # inherit bignum from $max_len
  my $ret_type = $options{'ret_type'} || 'accepting';
  my @counts = ($zero) x $self->num_states;

  ### @accepting
  # ### @counts
  foreach my $state (0) {   # starting
    $counts[$state]++;
  }

  my @ret;
  if ($ret_type eq 'rows') {
    @ret = map {[]} 0 .. $#counts;
  }

  foreach my $k (0 .. $max_len) {
    # ### at: "k=$k  ".join(',',map{$_//'_'}@counts)." total ".sum(0,map{$_//0}@counts)." accepting ".sum(0,map{$counts[$_]//0}@accepting)

    {
      my $accepting_count = $zero;
      foreach my $state (@accepting) {
        if ($counts[$state]) {
          $accepting_count += $counts[$state];
        }
      }
      if ($ret_type eq 'accepting') {
        push @ret, $accepting_count;
      } elsif ($ret_type eq 'columns') {
        push @ret, \@counts;
      } elsif ($ret_type eq 'rows') {
        foreach my $i (0 .. $#counts) {
          push @{$ret[$i]}, $counts[$i];
        }
      }
    }
    last if $k == $max_len;

    my @new_counts = ($zero) x $self->num_states;
    foreach my $arc (@{$self->{'arcs'}}) {
      my ($from_state,$to_state,$symbol) = @$arc;
      if (my $from_count = $counts[$from_state]) {
        ### add: "$from_state,$to_state,$symbol $from_count"
        $new_counts[$to_state] += $from_count;
      }
    }
    @counts = @new_counts;
  }
  return @ret;
}

1;
__END__
