#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Test of fBpred based on contiguous surrounding blob crossing lines.
# Bigger than the hullstouch.

use 5.006;
use strict;
use warnings;
use Test::More;

use FindBin;
use lib $FindBin::Bin;
use MyDragon ':all';

use MyTestHelpers;
BEGIN { MyTestHelpers::nowarnings(); }

plan tests => 243;

# uncomment this to run the ### lines
# use Smart::Comments;


#------------------------------------------------------------------------------
# fBpred by surrounding

# check that the expand(0) and expand(1) of $conf have all of $conf as subsets
sub ok_conf_expand_within_self {
  my ($conf) = @_;
  foreach my $bit (0,1) {
    ok($conf->is_subset( $conf->expand($bit) ),
       "$conf->{'name'}, bit=$bit");
  }
}

ok_conf_expand_within_self($surround_segs_inner);
ok_conf_expand_within_self($surround_segs_all);


#------------------------------------------------------------------------------
# surround state counts, not minimized

&fBpred_surround_init(verbose=>1);
### @fBpred_surround_state_enclosed_lists
is($fRpred_surround_flat->num_states, 184);  # as stated in dragon.tex
is($fLpred_surround_flat->num_states, 184);
is($fBpred_surround_flat->num_states, 184);

# as stated in past revisions of dragon.tex
# no eventually accepting
is (scalar(@{$fBpred_surround_state_enclosed_lists[0]}), 53);
is (scalar(@{$fBpred_surround_state_enclosed_lists[1]}), 53);
is (scalar(@{$fBpred_surround_state_enclosed_lists[2]}), 2); 

is($fNonRpred_surround_flat->MyFLAT::num_accepting, 77); # stated in dragon.tex
is($fNonLpred_surround_flat->MyFLAT::num_accepting, 76); # incl eventually
is($fNonBpred_surround_flat->MyFLAT::num_accepting, 22);

{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_surround_state_table,
     name => 'fNonRpred_surround',
     accepting_list => $fBpred_surround_state_enclosed_lists[0],
    );
  is ($f->MyFLAT::num_accepting, 53);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 3);
  is (scalar(@{$info{'states'}}), 77-53);
}
{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_surround_state_table,
     name => 'fNonRpred_surround',
     accepting_list => $fBpred_surround_state_enclosed_lists[1],
    );
  is ($f->MyFLAT::num_accepting, 53);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 3);
  is (scalar(@{$info{'states'}}), 76-53);
}
{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_surround_state_table,
     name => 'fNonRpred_surround',
     accepting_list => $fBpred_surround_state_enclosed_lists[2],
    );
  is ($f->MyFLAT::num_accepting, 2);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 3);
  is (scalar(@{$info{'states'}}), 22-2);
}

#------------------------------------------------------------------------------
# fRpred_flat() equality of surround

is(fRpred_flat()->num_states, 30);  # as stated in dragon.tex
is(fLpred_flat()->num_states, 32);

ok(fRpred_flat()->equals($fRpred_surround_flat),
    'fNonRpred_flat() equals fNonRpred_surround_flat');
ok(fLpred_flat()->equals($fLpred_surround_flat),
    'fNonLpred_flat() equals fNonLpred_surround_flat');
ok(fBpred_flat()->equals($fBpred_surround_flat),
    'fNonBpred_flat() equals fNonBpred_surround_flat');
ok(fNonRpred_flat()->equals($fNonRpred_surround_flat));
ok(fNonLpred_flat()->equals($fNonLpred_surround_flat));
ok(fNonBpred_flat()->equals($fNonBpred_surround_flat));

ok( fNonRpred_flat()
    ->equals($fNonRpred_surround_flat->MyFLAT::minimize(verbose=>1)) );
ok( fNonLpred_flat()
    ->equals($fNonLpred_surround_flat->MyFLAT::minimize(verbose=>1)) );
ok( fNonBpred_flat()
    ->equals($fNonBpred_surround_flat->MyFLAT::minimize(verbose=>1)) );


#------------------------------------------------------------------------------
# cases extras are on the boundary so are subset of fRpred, etc

fBpred_surround_init();
foreach my $extra (fRpred_cases_extra_list()) {
  ok($extra->is_subset_of($fRpred_surround_flat));
}
foreach my $extra (fLpred_cases_extra_list()) {
  ok($extra->is_subset_of($fLpred_surround_flat));
}
foreach my $extra (fRpred_cases_extra_list(), fLpred_cases_extra_list()) {
  ok($extra->is_subset_of($fBpred_surround_flat));
}

# Extras subtract from NonBpred_k_flat().
# Extra is bigger than NonBpred but has a non-empty intersection so that
# it does subtract.
foreach my $extra (fRpred_cases_extra_list(), fLpred_cases_extra_list()) {
  my $int = NonBpred_k_flat()->intersect($extra);
  ok(! $int->is_empty);
}

# anything with Spred at one end of segment is fBpred,
# including prefixes of Spred as they are not yet proven enclosed
ok (Spred_k_flat()->is_subset_of($fBpred_surround_flat));
ok (Spred_k_flat()
    ->MyFLAT::prefix
    ->is_subset_of($fBpred_surround_flat));



#------------------------------------------------------------------------------
# fBpred surround which are fully enclosed or eventually fully enclosed
#  =
# those with all of $triple_15ths_union_segs

{
  my $conf = $triple_15ths_union_segs;
  {
    my $bit = 1;
    ok(! $conf->is_subset( $conf->expand($bit) ),
       "does not expand to within self $conf->{'name'}, bit=$bit");
  }

  my $at_end = MyDragonConf->new
    (name => 'e at end',
     side => 0,
     segs => [ "2,0,2" ]);

  $conf = $conf->union($at_end)->set_name('triples union plus end');
  foreach my $bit (0,1) {
    ok($conf->is_subset( $conf->expand($bit) ),
       "expands to within self $conf->{'name'}, bit=$bit");
  }
}

fBpred_surround_init();

foreach my $state (0 .. $#fBpred_surround_state_table) {
  my $conf = $fBpred_surround_state_to_conf[$state];
  my $by_table   = $fNonBpred_surround_flat->is_accepting($state) ? 1 : 0;
  my $by_triples = $triple_15ths_union_segs->is_subset($conf) ? 1 : 0;

  if ($by_table != $by_triples) {
    $triple_15ths_union_segs->print;
    $conf->print;
    $triple_15ths_union_segs->difference($conf)->print;
  }
  is ($by_triples, $by_table);
}

#------------------------------------------------------------------------------
exit 0;
