#!/usr/bin/perl -w

# Copyright 2020, 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use warnings;
use 5.006;
use FLAT;
use List::Util 'max';
use Test::More;

use FindBin;
use lib $FindBin::Bin;
use MyFLAT ':all';
use MyDragon ':all';

use MyTestHelpers;
BEGIN { MyTestHelpers::nowarnings(); }

plan tests => 780;


#------------------------------------------------------------------------------
# @fVisits_adjacent_list

{
  my $conf = MyDragonConf->new
    (side => 0,
     segs => [ '0,0,0', grep {defined} @fVisits_adjacent_list ]);
  ok ($conf->equals($hullstouch_segs_all),
      'fVisits_adjacent_list all of hullstouch_segs_all');
  is ($#fVisits_adjacent_list,
      scalar(keys %fVisits_seg_to_index),
      'fVisits_adjacent_list no duplicates');
}


#------------------------------------------------------------------------------
# fVisits_expand_bit()

{
  #    \  s1
  #    /
  #   O  *
  #   0\/1
  is_deeply([MyDragon::fVisits_expand_bit(1,0)],  [1], 'fVisits_expand_bit');
  is_deeply([MyDragon::fVisits_expand_bit(1,1)],  [6]);

  #   O  *
  #    \/ \  s8
  #       /
  is_deeply([MyDragon::fVisits_expand_bit(8,0)],  [11]);
  is_deeply([MyDragon::fVisits_expand_bit(8,1)],  [1]);

  #      O  *
  #  s5 / \/ 
  #     \    
  is_deeply([MyDragon::fVisits_expand_bit(5,0)],  [5,6]);
  is_deeply([MyDragon::fVisits_expand_bit(5,1)],  []);
}

{
  is(MyDragon::fVisits_state_num(0,0), 0);
  is(MyDragon::fVisits_state_num(11,0), 11);
  is(MyDragon::fVisits_state_num(1,1), 12);
  is(MyDragon::fVisits_state_num(11,1), 22);

  my $f = MyDragon::fVisits_adjacent_NFA();
  is($f->num_states,1+11+11);
}


#------------------------------------------------------------------------------
# fVisits_adjacent_NFA()

# Reverse
#                  * s1  3/15
#                  ^
#                  |
#    1/15 s2  *<---*--->*  m 11/15
#
#           13---12
#            |
#    19,15-14,22-23    double-visited even 
#            |    |
#          21,25-24    
#
# dragon_fpoint( (14-11/15)/32 ) == 11/24 - 7/24*I
# dragon_fpoint( (14+ 3/15)/32 ) == 11/24 - 7/24*I
# dragon_fpoint( (22- 1/15)/32 ) == 11/24 - 7/24*I
#
# dragon_fpoint( (14+11/15)/32 ) == 13/24 - 7/24*I
# dragon_fpoint( (22- 3/15)/32 ) == 13/24 - 7/24*I
# dragon_fpoint( (22+ 1/15)/32 ) == 13/24 - 7/24*I

{
  fBpred_hullstouch_init();
  my $f = $fBpred_hullstouch_flat->clone
    ->MyFLAT::as_nfa;
  $f->unset_accepting($f->get_states);
  diag "f states ",$f->num_states;

  my $offset = $f->num_states;
  my $v = MyDragon::fVisits_adjacent_NFA();
  diag "v states ",$v->num_states;
  $f->add_states($v->num_states);
  diag "f joined states ",$f->num_states;
  foreach my $state ($v->get_accepting) {
    $f->set_accepting ($state+$offset);
  }
  foreach my $state ($v->get_states) {
    foreach my $bit (0,1) {
      foreach my $to_state ($v->successors($state, $bit)) {
        $f->add_transition($state+$offset, $to_state+$offset, $bit);
      }
    }
  }
  foreach my $state (0 .. $offset-1) {
    my $conf = $fBpred_hullstouch_state_to_conf[$state];
    my $count_epsilons = 0;
    foreach my $seg ($conf->segs_list) {
      next if $seg eq '0,0,0';
      my $s = $fVisits_seg_to_index{$seg} // die "oops unknown seg";
      my $v_state = fVisits_state_num($s,$conf->{'side'});
      $f->add_transition($state, $v_state+$offset, ''); # epsilon
      $count_epsilons++;
    }
    # diag "epsilons $count_epsilons";
  }
  $f = $f->as_dfa;
  $f = $f->MyFLAT::minimize(verbose=>1);
  # $f->MyFLAT::view;
}
{
  # fVisits_adjacent_NFA intersection of 1 vs eventually fBpred

  my $subcurve_fBpred = FLAT_any_binary()->MyFLAT::concat(fBpred_flat())
    ->MyFLAT::set_name('sub-curve fBpred')
    # ->MyFLAT::minimize(verbose=>1)
    ;
  diag "subcurve_fBpred num states ", $subcurve_fBpred->num_states;

  my $f = $subcurve_fBpred->difference(fBpred_flat())
    # ->MyFLAT::minimize(verbose=>1)
    ;
  diag "subcurve_fBpred difference num states ", $f->num_states;

  $f = $f->complement;
  $f = MyFLAT::eventually_accepting($f);
  $f = $f->complement;

  ok ($f->equals(fNonBpred_flat()),
      'equal');

  # $f = $f->MyFLAT::minimize(verbose=>1);
  # $f->MyFLAT::view;
}
{
  # fVisits_adjacent_NFA intersection of 1 vs eventually fBpred

  my $subcurve_fBpred = FLAT_any_binary()->MyFLAT::concat(fBpred_flat())
    ->MyFLAT::set_name('sub-curve fBpred')
    # ->MyFLAT::minimize(verbose=>1)
    ;
  diag "subcurve num states ",$subcurve_fBpred->num_states;

  my $f = fNonBpred_flat()->intersect($subcurve_fBpred)
    # ->MyFLAT::minimize(verbose=>1)
    ;
  diag "intersect num states ",$f->num_states;
  ok ($f->equals(fNonBpred_flat()));

  # $f = $f->complement;
  # $f = MyFLAT::eventually_accepting($f);
  # $f = $f->complement;
  # $f = $f->MyFLAT::minimize(verbose=>1);
  # 
  # $f->MyFLAT::view;
}


{
  # fVisits_adjacent_NFA intersection of 2

  # $triple_15ths_segs[1]->print;
  # $triple_15ths_segs[3]->print;
  # $triple_15ths_segs[11]->print;

  my %pair_to_exact_bit = ('1,2' => 0,
                           '1,5' => 0,
                           '2,5' => 0,
                           '8,9' => 1,
                           '8,10' => 1,
                           '9,10' => 1);
  my @bit_to_f = (FLAT_zeros(), FLAT_ones());

  # state machines matching digits of 7th or 15th
  my @seventh_to_f   = map { MyFLAT::fraction_digits($_,7, radix => 2) } 0..6;
  my @fifteenth_to_f = map { MyFLAT::fraction_digits($_,15, radix => 2) } 0..14;

  my %pair_to_15th_frac;
  my %pair_to_15th_f;
  foreach my $m (1 .. 15) {
    if ($m % 5 == 0) {
      ok (! defined $triple_15ths_segs[$m], '15ths not mod 5');
      next;
    }
    my $conf = $triple_15ths_segs[$m];
    # $conf->print;

    my @others;
    my @segs = $conf->segs_list;
    next if @segs < 3;
    foreach my $seg ($conf->segs_list) {
      next if $seg eq '0,0,0';  # not m itself
      my $s = $fVisits_seg_to_index{$seg} // die "oops unknown seg";
      push @others, $s;
    }
    @others = sort @others;
    is (scalar(@others), 2, '15ths segs 2 others');
    my $key = join(',',@others);
    $pair_to_15th_frac{$key} = $m;
  }

  my %pair_to_7th_frac;
  my %pair_to_7th_f;
  foreach my $m (1 .. 6) {
    my $conf = $triple_7ths_segs[$m];
    # $conf->print;

    my @others;
    my @segs = $conf->segs_list;
    next if @segs < 3;
    foreach my $seg ($conf->segs_list) {
      next if $seg eq '0,0,0';
      my $s = $fVisits_seg_to_index{$seg};
      ok (defined $s, '7th seg is among fVisits segs');
      push @others, $s;
    }
    @others = sort @others;
    is (scalar(@others), 2, '7ths segs 2 others');
    my $key = join(',',@others);
    $pair_to_7th_frac{$key} = $m;
    $pair_to_7th_f{$key} = MyFLAT::fraction_digits($m,7, radix => 2);
  }

  foreach my $rev (0,1) {
    foreach my $a (1 .. 10) {
      my $f = fVisits_adjacent_DFA($a,$rev);
      foreach my $b ($a+1 .. 11) {
        my $f = $f->intersect(fVisits_adjacent_DFA($b,$rev))
          ->MyFLAT::set_name("fVisits intersect $a,$b rev=$rev");
        next if $f->is_finite;

        my $key = "$a,$b";
        my $frac_7th  = $pair_to_7th_frac{$key};
        my $frac_15th = $pair_to_15th_frac{$key};
        my $exact_bit = $pair_to_exact_bit{$key};
        if ($rev) {
          if (defined $frac_7th)  { $frac_7th  = 7  - $frac_7th; }
          if (defined $frac_15th) { $frac_15th = 15 - $frac_15th; }
          if (defined $exact_bit) { $exact_bit = 1  - $exact_bit; }
        }
        my $f_7th   = (defined $frac_7th ? $seventh_to_f[$frac_7th] : undef);
        my $f_15th  = (defined $frac_15th ? $fifteenth_to_f[$frac_15th] :undef);
        my $f_exact = (defined $exact_bit ? $bit_to_f[$exact_bit] : undef);
        $exact_bit //= 'none';
        $frac_7th //= 'none';
        $frac_15th //= 'none';

        # # debug print:
        # if ($key eq '1,5' && $rev) {
        #   $f = $f->MyFLAT::minimize(verbose=>1);
        #   $f->MyFLAT::view;
        #   if (defined $f_15th) { $f_15th->MyFLAT::view; }
        #   # MyFLAT::FLAT_show_breadth($f, 12);
        # }

        if (defined $f_7th) {
          ok ($f_7th->is_subset_of($f),
              "fVisits intersection $key rev=$rev has $frac_7th/7 as subset");
        }
        if (defined $f_15th) {
          ok ($f_15th->is_subset_of($f),
              "fVisits intersection $key rev=$rev has $frac_15th/15 as subset");
        }
        if (defined $f_exact) {
          ok ($f_exact->is_subset_of($f),
              "fVisits intersection $key rev=$rev has exact $exact_bit as subset");
        }
        if (defined $f_7th) { $f = $f->difference($f_7th); }
        if (defined $f_15th) { $f = $f->difference($f_15th); }
        if (defined $f_exact) { $f = $f->difference($f_exact); }
        ok ($f->is_finite,
            "fVisits intersection $key rev=$rev finite other than exact $exact_bit and $frac_7th/7 and $frac_15th/15");
      }
    }
  }
}

{
  # fVisits_adjacent_NFA intersection of 3, all end 0s or 1s
  # so no fVisits >= 4, only exact Dpred doubles

  foreach my $rev (0,1) {
    foreach my $a (1 .. 9) {
      my $f = fVisits_adjacent_DFA($a,$rev);
      foreach my $b ($a+1 .. 10) {
        my $f = $f->intersect(fVisits_adjacent_DFA($b,$rev))
          ->MyFLAT::minimize(verbose=>1);
        next if $f->is_finite;

        foreach my $c ($b+1 .. 11) {
          my $f = $f->intersect(fVisits_adjacent_DFA($c,$rev));
          next if $f->is_finite;

          $f = $f->MyFLAT::minimize(verbose=>1);
          # $f->MyFLAT::view;
          ok ("$a,$b,$c" eq '1,2,5' || "$a,$b,$c" eq '8,9,10',
              'fVisits intersection 3 is a double-visited');

          if ("$a,$b,$c" eq '1,2,5') {
            my $reps    = ($rev ? FLAT_ones() : FLAT_zeros());
            my $repname = $reps->{'name'};
            ok ($reps->is_subset_of($f),
                "fVisits intersection 1,2,5 contains $repname...");
            ok ($f->difference($reps)->is_finite,
                "fVisits intersection 1,2,5 finite except for $repname...");
          } elsif ("$a,$b,$c" eq '8,9,10') {
            my $reps    = ($rev ? FLAT_zeros() : FLAT_ones());
            my $repname = $reps->{'name'};
            ok ($reps->is_subset_of($f),
                "fVisits intersection 8,9,10 contains $repname...");
            ok ($f->difference(FLAT_ones())->is_finite,
                "fVisits intersection 8,9,10 finite except for $repname...");
          }

          $f = $f->complement;
          $f = MyFLAT::eventually_accepting($f);
          $f = $f->complement;
          # $f->MyFLAT::view;
          # MyFLAT::FLAT_show_breadth($f, 8);

          if ("$a,$b,$c" eq '1,2,5') {
            my $reps    = ($rev ? FLAT_ones() : FLAT_zeros());
            my $repname = $reps->{'name'};
            ok ($f->equals($reps),
                "fVisits intersection 1,2,5 want $repname...");
          } elsif ("$a,$b,$c" eq '8,9,10') {
            my $reps    = ($rev ? FLAT_zeros() : FLAT_ones());
            my $repname = $reps->{'name'};
            ok ($f->equals($reps),
                "fVisits intersection 8,9,10 want $repname...");
          }
        }
      }
    }
  }
}

{
  # fVisits_adjacent_NFA intersection of 4, all are finite,
  # so no fVisits >= 5

  my $max_length = 0;

  foreach my $rev (0,1) {
    foreach my $a (1 .. 8) {
      my $f = fVisits_adjacent_DFA($a,$rev);
      foreach my $b ($a+1 .. 9) {
        my $f = $f->intersect(fVisits_adjacent_DFA($b,$rev))
          ->MyFLAT::minimize(verbose=>1);

        foreach my $c ($b+1 .. 10) {
          my $f = $f->intersect(fVisits_adjacent_DFA($c,$rev))
            ->MyFLAT::minimize(verbose=>1);

          foreach my $d ($c+1 .. 11) {
            my $f = $f->intersect(fVisits_adjacent_DFA($d,$rev));

            # $f->MyFLAT::view(synchronous=>1);
            ok($f->is_finite,
               "$a $b $c $d intersection finite");
            my $length = $f->MyFLAT::finite_max_length;
            my @counts = MyFLAT::FLAT_count_contains($f, 10);
            diag "$a $b $c $d length $length counts ", join(',',@counts);
            $max_length = max($max_length, $length);

            foreach my $i (reverse 0 .. $#counts) {
              if ($counts[$i]) {
                $length == $i or die "oops";
                last;
              }
            }
          }
        }
      }
    }
  }
  is($max_length, 5,
     'fVisits longest finite from intersection 4 others');
}
{
  # fBpred made from fVisits_adjacent_NFA

  # NFA with all s1..s11 as starts, is union of starting at any of them
  my $f = MyDragon::fVisits_adjacent_NFA()->clone;
  $f->set_starting(map {fVisits_state_num($_,0)} 1..11);
  $f = $f->as_dfa;

  # The unminimized state count after as_dfa() may depend how NFA->DFA is
  # implemented, so don't make it a test as such.
  # is ($f->num_states, 91,
  #     'fVisits_adjacent_NFA() any start DFA num states');

  ok(fBpred_flat()->equals($f),
     'fVisits_adjacent_NFA() any start = fBpred_flat()');
  # FLAT_check_is_equal($f, fBpred_flat());
}

# {
#   my $u;
#   foreach my $s (1 .. 11) {
#     my $f = MyDragon::fVisits_adjacent_DFA($s,0);
#     $f = $f->complement;
#     $f = MyFLAT::eventually_accepting($f);
#     $f = $f->complement;
#     $f = $f->MyFLAT::minimize(verbose=>1);
#     if (defined $u) {
#       $u = $u->union($f)->MyFLAT::set_name("to s$s");
#       $u = $u->MyFLAT::minimize(verbose=>1);
#     } else {
#       $u = $f;
#     }
#   }
#   $u = $u->complement;
#   $u = MyFLAT::eventually_accepting($u);
#   $u = $u->complement;
#   $u = $u->MyFLAT::minimize(verbose=>1);
# 
#   # $u->MyFLAT::view;
#   ok(fBpred_flat()->equals($u),
#      'fVisits_adjacent_DFA() union is fBpred_flat()');
#   FLAT_check_is_equal($u, fBpred_flat());
# }


#------------------------------------------------------------------------------
exit 0;
