#!/usr/bin/perl -w

# Copyright 2019, 2020, 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use List::Util 'uniqstr';
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon '@dir4_to_dx','@dir4_to_dy','xy_to_node_name';
use MyPrintwrap ':all';
use MyLineSegments;
$| = 1;


my $type = $ARGV[0] || '3even';

print "  % generated by tools/$FindBin::Script $type\n";

my %type_to_info
  = ('3even' => { segs => ['0,0,0','0,0,3',
                           ,'1,-1,2',
                           '-1,-1,0',
                           '0,-2,0','0,-2,1','0,-2,2','0,-2,3',
                           '1,-3,0','1,-3,1','1,-3,2','1,-3,3',
                          ],
                  blanks => {'1,0' => 1,
                             '0,-3' => 1,
                             '-1,-1' => 1,
                             '-1,-2' => 1,
                             '2,-3' => 1,
                             '1,-4' => 1,
                            },
                  slope_x => 1,
                  slope_y => -3,
                  BP_dx => -1,
                  BP_dy => 1,
                  compare_reps => 5,
                },
     '3odd'  => { segs => ['0,0,0', '0,0,1',
                           '1,1,2','1,1,3',
                           '-1,1,0',
                           '0,2,0','0,2,1','0,2,2','0,2,3',
                           '1,3,2','1,3,3',
                           '0,4,3','-1,3,0',
                          ],
                  blanks => {'0,0' => 1,
                             '1,3' => 1,
                             '-1,1' => 1,
                             '-1,2' => 1,
                             '-1,3' => 1,
                             '0,4' => 1,
                            },
                  harpoons => { '1,1,3' => '',
                                '0,2,1' => '',
                              },
                  slope_x => -1,
                  slope_y => 3,
                  BP_dx => 2,
                  BP_dy => -2,
                  compare_reps => 4,
                },

     '4odd'  => { segs => ['0,0,0', '0,0,3',
                           '1,-1,1','1,-1,2',
                           '0,-2,0','0,-2,1','0,-2,2','0,-2,3',
                           '-1,-3,0','-1,-3,1','-1,-3,2','-1,-3,3',
                           '0,-4,1','0,-4,2',
                           '-2,-4,0',
                           '-1,-5,1',
                          ],
                  blanks => {'0,0' => 1,
                             '1,-2' => 1,
                             '-1,-2' => 1,
                             '0,-4' => 1,
                            },
                  slope_x => -2,
                  slope_y => -4,
                  BP_dx => 3,
                  BP_dy => 1,
                  compare_reps => 4,
                },
     '4even' => { segs => ['0,0,0', '0,0,1',
                           '1,1,0','1,1,1','1,1,2','1,1,3',
                           '2,2,1','2,2,2',
                           '1,3,0','1,3,1',
                           '2,4,0','2,4,1','2,4,2','2,4,3',
                          ],
                  blanks => {'1,0' => 1,
                             '2,1' => 1,
                             '2,2' => 1,
                             '1,4' => 1,
                             '3,4' => 1,
                             '2,5' => 1,
                            },
                  slope_x => 2,
                  slope_y => 4,
                  BP_dx => 0,
                  BP_dy => -2,
                  compare_reps => 4,
                },
    );

my $info = $type_to_info{$type} || die "Unknown type: $type";
{
  # no duplicates in $info->{'segs'}
  my @uniq = uniqstr(@{$info->{'segs'}});
  $#{$info->{'segs'}} == $#uniq or die;
}
my $conf = MyDragonConf->new (segs => $info->{'segs'},
                              side => 0);
{
  my $str = $conf->print_to_str;
  $str =~ s/^/  %    /mg;
  print $str;
  print "  %\n";
}
foreach my $seg (keys %{$info->{'harpoons'}}) {
  $conf->contains($seg) or die "harpoon no such segment $seg";
}

my @dots;
{
  my %seen;
  my %unused_blanks = %{$info->{'blanks'}};
  $Printwrap = 2;
  foreach my $seg (sort $conf->segs_list) {
    my ($x,$y,$dir) = split /,/, $seg;
    foreach (0, 1) {
      my $name = xy_to_node_name($x,$y);
      unless ($seen{"$x,$y"}++) {
        my $blank = '';
        if ($info->{'blanks'}->{"$x,$y"}) {
          $blank = ' blank';
          delete $unused_blanks{"$x,$y"};
        } else {
          push @dots, (4*$x).','.(4*$y);
        }
        printwrap(" \\node ($name) at ($x,$y) [my$blank dot];");
      }
      $x += $dir4_to_dx[$dir];
      $y += $dir4_to_dy[$dir];
    }
  }
  @dots = sort @dots;
  
  if (%unused_blanks) {
    die "unused blanks ",join(' ',sort keys %unused_blanks);
  }
}
printwrap_newline();

foreach my $seg (sort $conf->segs_list) {
  my ($x,$y,$dir) = split /,/, $seg;
  my ($x2,$y2) = ($x+$dir4_to_dx[$dir], $y+$dir4_to_dy[$dir]);
  my $from = xy_to_node_name($x,$y);
  my $to   = xy_to_node_name($x2,$y2);
  my $harpoon = $info->{'harpoons'}->{$seg};
  my $arrow = (defined $harpoon ? "-{>[harpoon$harpoon,scale width=1.2]}" : '->');
  printwrap(" \\draw[$arrow] ($from)--($to);");
}
printwrap_newline();

my $exp = $conf;
foreach (1..4) { $exp = $exp->expand(0); }
$exp = $exp->rotate(2);

print "  \\begin{scope}[shift=(right-pos),scale=1/4]\n";
{
  my @block_positions;
  my ($cmpstart_x,$cmpstart_y) = ($info->{'BP_dx'}, $info->{'BP_dy'});
  my ($cmp_x,$cmp_y) = ($cmpstart_x,$cmpstart_y);
  my $sx = $info->{'slope_x'};
  my $sy = $info->{'slope_y'};
  my $compare_reps = $info->{'compare_reps'};
  foreach (1 .. $compare_reps) {
    my $sub = $conf->translate ($cmp_x,$cmp_y);
    unless ($sub->is_subset($exp)) {
      print "Missing segments on expansion translate $cmp_x,$cmp_y:\n";
      $sub->difference($exp)->print;
      die;
    }
    $cmp_x += $sx;
    $cmp_y += $sy;
  }
  print "  % compare segments at $compare_reps places cmp $cmp_x, $cmp_y\n";
  print "  \\coordinate (compare-start) at ($cmpstart_x,$cmpstart_y);\n";
  print "  \\coordinate (compare) at ($cmp_x,$cmp_y);\n";
}

foreach my $xy (@dots) {
  printwrap(" \\node at ($xy) [my small dot];");
}

my $segments = MyLineSegments->new;
$Printwrap = 2;
foreach my $seg (sort $exp->segs_list) {
  my ($x,$y,$dir) = split /,/, $seg;
  my ($x2,$y2) = ($x+$dir4_to_dx[$dir], $y+$dir4_to_dy[$dir]);
  $segments->add($x,$y, $x2,$y2);
}
printwrap_newline();

$Printwrap = 2;
printwrap("\\draw ");
$Printwrap += 2;
$segments->printwrap;
printwrap(";\n");

print "  \\end{scope}\n";

print "  % end generated\n";
exit 0;
