#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use List::Util 'min';
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'xy_rotate_plus90','xy_rotate_minus90','BlobStart_xy','BlobEnd_xy',
  'transform_to_horizontal';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 12;

print "  % generated by tools/$FindBin::Script\n";
my $path = Math::PlanePath::DragonCurve->new;

my ($n_lo, $n_hi) = $path->level_to_n_range($level);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
my ($x_join,$y_join) = $path->n_to_xy($n_hi/2);
### level: "$level  n_hi=$n_hi"

sub transform {
  my ($x,$y) = @_;
  return transform_to_horizontal($x,$y, $level);
}

my $segments = MyLineSegments->new;
my $y_min = 0;

my $x_middle = $x_hi/2;
my $y_middle = $y_hi/2;

sub n_is_right_boundary {
  my ($n) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to right side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! defined $path->xyxy_to_n_either($x1,$y1, $x1s,$y1s)
          || ! defined $path->xyxy_to_n_either($x2,$y2, $x2s,$y2s)
          || ! defined $path->xyxy_to_n_either($x1s,$y1s, $x2s,$y2s));
}

my $Rnear_norm = $n_hi;
my $Rnear_x = 0;
my $Rnear_y = 0;
my $RnearN = 0;
foreach my $n ($n_lo .. $n_hi-1) {
  my ($x,$y) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  $segments->add($x,$y, $x2,$y2);

  if (n_is_right_boundary($n)) {
    my $norm = ($x-$x_middle)**2 + ($y-$y_middle)**2;
    if ($norm < $Rnear_norm) {
      $Rnear_norm = $norm;
      $Rnear_x = $x;
      $Rnear_y = $y;
      $RnearN = $n;
    }
  }

  ($x,$y) = transform($x,$y);
  ($x2,$y2) = transform($x2,$y2);
  $y_min = min($y_min, $y);
}

$Printwrap = 2;
printwrap("\\draw[my dragon] ");
$Printwrap += 2;
$segments->printwrap (transform => \&transform);
printwrap(";\n");

sub blob_crossing_line {
  my ($sx,$sy, $tx,$ty) = @_;
  ($sx,$sy) = transform($sx,$sy);
  ($tx,$ty) = transform($tx,$ty);
  my $len = ($sx-$tx)**2 + ($sy-$ty)**2;
  my $style = ($len <= 2 ? "my zero line" : "my line");
  printwrap(" \\draw[$style] ($sx,$sy)--($tx,$ty);");
}

{
  $Printwrap = 2;

  # start blobs
  foreach my $k (4 .. $level-1) {
    my ($sx,$sy) = BlobStart_xy($k);
    my ($tx,$ty) = BlobEnd_xy($k);
    blob_crossing_line ($sx,$sy, $tx,$ty);

    # same for second sub-curve
    ($sx,$sy) = xy_rotate_minus90($sx,$sy);
    ($tx,$ty) = xy_rotate_minus90($tx,$ty);
    $sx += $x_hi; $sy += $y_hi;
    $tx += $x_hi; $ty += $y_hi;
    blob_crossing_line ($sx,$sy, $tx,$ty);
  }

  # end blobs, back from join
  foreach my $k (4 .. $level-3) {
    my ($sx,$sy) = BlobStart_xy($k);
    my ($tx,$ty) = BlobEnd_xy($k);
    ($sx,$sy) = xy_rotate_minus90($sx,$sy);
    ($tx,$ty) = xy_rotate_minus90($tx,$ty);
    $sx += $x_join;
    $sy += $y_join;
    $tx += $x_join;
    $ty += $y_join;
    blob_crossing_line ($sx,$sy, $tx,$ty);

    # same for second sub-curve
    ($sx,$sy) = xy_rotate_minus90($sx,$sy);
    ($tx,$ty) = xy_rotate_minus90($tx,$ty);
    $sx += $x_hi; $sy += $y_hi;
    $tx += $x_hi; $ty += $y_hi;
    blob_crossing_line ($sx,$sy, $tx,$ty);
  }
}
printwrap_newline();

{
  my ($x,$y) = transform(0,0);
  print "  \\node (start) at ($x,$y) [my dot];\n";
}
{
  my ($x,$y) = transform($x_hi/2,$y_hi/2);
  print "  \\node (middle) at ($x,$y) [my dot];\n";
}
{
  my ($x,$y) = transform($x_hi,$y_hi);
  print "  \\node (end) at ($x,$y) [my dot];\n";
}
{
  my ($x,$y) = transform($x_join,$y_join);
  print "  \\node (join) at ($x,$y) [my dot];\n";
}
{
  my ($x,$y) = transform($Rnear_x,$Rnear_y);
  print "  \\node (Rnear) at ($x,$y) [my dot];\n";
}

{
  my $k = $level - 2;
  my ($sx,$sy) = BlobStart_xy($k);
  ($sx,$sy) = xy_rotate_plus90($sx,$sy);
  my ($tx,$ty) = ($x_hi-$sx,$y_hi-$sy);

  ($sx,$sy) = transform($sx,$sy);
  ($tx,$ty) = transform($tx,$ty);
  print "  \\coordinate (middle-start) at  ($sx,$sy);\n";
  print "  \\coordinate (middle-end) at  ($tx,$ty);\n";
}

print "  \\coordinate (bottom) at (0,$y_min);\n";
print "% GP-Test  Rnear($level)^2 == $Rnear_norm\n";
print "% GP-Test  Rnear_z($level) + 1/2*b^$level == $Rnear_x + I*$Rnear_y\n";

print "  % end generated\n";
exit 0;
