#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use Lingua::EN::Numbers;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()
use Tie::IxHash;

use FindBin;
use lib $FindBin::Bin;
use MyDragon ':all';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 12;
my $mod = 0;
if (@ARGV) { $mod = shift @ARGV; }
print "  % generated by tools/$FindBin::Script $mod\n";

my $path = Math::PlanePath::DragonCurve->new;

my $k = $level + $mod;
my ($n_lo, $n_hi) = $path->level_to_n_range($k);

my $name = ucfirst(Lingua::EN::Numbers::num2en($mod));
print "  \\newcommand{\\MyLevel$name}{$k}\n";
my $scale = ($mod & 1 ? "scale={1/sqrt(2)}," : "");
print "  \\begin{scope}[${scale}shift=(pos-$mod)]\n";

my $bridge_n_end = BlobN($k+1);
my $bridge_n = $bridge_n_end - 3;

my $near_n = BlobN($k+1);
if ($mod == 0) { $near_n -= 2; }
if ($mod == 1) { $near_n -= 3; }
if ($mod == 2) { $near_n -= 2; }
if ($mod == 3) { $near_n -= 1; }

my ($near_x,$near_y) = $path->n_to_xy($near_n);

my ($near_x_transformed, $near_y_transformed)
  = transform_to_horizontal($near_x,$near_y, $k);
### $near_x_transformed

my $transform = sub {
  my ($x,$y) = @_;
  ($x,$y) = transform_to_horizontal($x,$y, $k);
  $x -= $near_x_transformed;
  return ($x,$y);
};

my $y_min;
my $y_max;
my $x_min;
my $x_max;
{
  my ($x,$y) = $transform->($near_x,$near_y);
  $x == 0 or die "oops near x transformed is $x";
  if ($mod == 0) {
    $x_min = $x - 3;
    $x_max = $x + 4;
    $y_min = $y - 4;
    $y_max = $y + 3;
  }
  if ($mod == 2) {
    $x_min = $x - 3;
    $x_max = $x + 4;
    $y_min = $y - 4;
    $y_max = $y + 3;
  }

  if ($mod == 1) {
    $x_min = $x - 5;
    $x_max = $x + 7;
    $y_min = $y - 5;
    $y_max = $y + 6;
  }
  if ($mod == 3) {
    $x_min = $x - 5;
    $x_max = $x + 5;
    $y_min = $y - 6;
    $y_max = $y + 5;
  }
}

my $n_is_right_boundary = sub {
  my ($n) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to right side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! defined $path->xyxy_to_n_either($x1,$y1, $x1s,$y1s)
          || ! defined $path->xyxy_to_n_either($x2,$y2, $x2s,$y2s)
          || ! defined $path->xyxy_to_n_either($x1s,$y1s, $x2s,$y2s));
};

my $clip = sub {
  my ($x1,$y1, $x2,$y2) = @_;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2) // return 0;
  if ($n > $n_hi) { return 0; }
  # $n_is_right_boundary->($n) || return 0;

  ($x1,$y1) = $transform->($x1,$y1);
  ($x2,$y2) = $transform->($x2,$y2);
  if ($x_min <= $x1 && $x1 <= $x_max
      && $y_min <= $y1 && $y1 <= $y_max
      && $x_min <= $x2 && $x2 <= $x_max
      && $y_min <= $y2 && $y2 <= $y_max) {
    return $n;
  }
  return 0;
};

my $segments = MyLineSegments->new;
my $bridge_segments = MyLineSegments->new;
foreach my $x ($near_x-15 .. $near_x+15) {
  foreach my $y ($near_y-15 .. $near_y+15) {
    foreach my $dx (0, 1) {
      my ($x2,$y2) = ($x+$dx,$y+(1-$dx));
      my $n = $clip->($x,$y, $x2,$y2) || next;
      if ($n >= $bridge_n && $n < $bridge_n_end) {
        $bridge_segments->add($x,$y, $x2,$y2);
      } else {
        $segments->add($x,$y, $x2,$y2);
      }
    }
  }
}

# foreach my $delta (1, -1) {
#   for (my $n = $near_n; ; $n += $delta) {
#     $n_is_right_boundary->($n) || next;
#
#     my ($x,$y) = $path->n_to_xy($n);
#     my ($x2,$y2) = $path->n_to_xy($n+1);
#     $clip->($x,$y, $x2,$y2) || last;
#     $segments->add($x,$y, $x2,$y2);
#   }
# }

$Printwrap = 4;
printwrap("\\draw[my grey]");
$Printwrap += 2;
$segments->printwrap (transform => $transform);
printwrap(";\n");

$Printwrap = 4;
printwrap("\\draw[my bridge]");
$Printwrap += 2;
$bridge_segments->printwrap (transform => $transform);
printwrap(";\n");

$Printwrap = 4;
{
  my ($x,$y) = $transform->($near_x,$near_y);
  printwrap("\\node (near-$mod) at ($x,$y) [my dot];\n");
}
if ($mod == 1) {
  my ($x,$y) = $path->n_to_xy($near_n + 2);
  ($x,$y) = $transform->($x,$y);
  printwrap("\\node (near-$mod-other) at ($x,$y) [my dot];\n");
}
print "  \\end{scope}\n";

print "  % end generated\n";
exit 0;
