#!/usr/bin/perl -w

# Copyright 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'BlobN','transform_to_horizontal','chamfer';
use MyLineSegments;
use MyPrintwrap ':all';


my $level = 8;
if (@ARGV) { $level = $ARGV[0]; }

print "    % generated by tools/$FindBin::Script $level\n";
print "    \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::DragonCurve->new;

my $n_blob_start = BlobN($level);
my $n_blob_end   = BlobN($level+1) - 3;
my $n_lo = $n_blob_start - 1;  # one segment extra each end
my $n_hi = $n_blob_end   + 1;

sub transform_rotate {
  my ($x,$y) = @_;
  return transform_to_horizontal($x,$y, $level + ($level&1));
}
my ($x_offset,$y_offset) = $path->n_to_xy($n_blob_start);
sub transform {
  my ($x,$y) = @_;
  return transform_rotate($x-$x_offset, $y-$y_offset);
}
{
  my ($x,$y) = transform($path->n_to_xy($n_blob_start));
  print "    \\coordinate (BlobStart) at ($x,$y);\n";
}
{
  my ($x,$y) = transform($path->n_to_xy($n_blob_end));
  print "    \\coordinate (BlobEnd) at ($x,$y);\n";
}
{
  my ($x,$y) = transform($path->n_to_xy($n_lo));
  print "    \\coordinate (start) at ($x,$y);\n";
}
{
  my ($x,$y) = transform($path->n_to_xy($n_hi));
  print "    \\coordinate (end) at ($x,$y);\n";
}

my $subcurve_2_lo = 2 * 2**($level-2);
my $subcurve_2_hi = 3 * 2**($level-2);
{
  my ($x,$y) = transform($path->n_to_xy($subcurve_2_lo));
  print "    \\coordinate (p2) at ($x,$y);\n";
}
{
  my ($x,$y) = transform($path->n_to_xy($subcurve_2_hi));
  print "    \\coordinate (p3) at ($x,$y);\n";
}

my $bridge_end   = $subcurve_2_lo + BlobN($level-1);
my $bridge_start = $bridge_end - 3;

my $mask = 3 << ($level-2);
foreach my $elem ([$n_lo,          $subcurve_2_lo, 'my pre line'],
                  [$subcurve_2_lo, $bridge_start,  'my mid line'],
                  [$bridge_start,  $bridge_end,    'my bridge line'],
                  [$bridge_end,    $subcurve_2_hi, 'my mid line'],
                  [$subcurve_2_hi, $n_hi,          'my post line']) {
  my ($start,$end,$style) = @$elem;
  $start <= $end or die "oops $start to $end for $style";
  next if $start == $end;

  $Printwrap = 4;
  printwrap("\\draw[$style] ");
  $Printwrap += 2;
  my $join = '';
  foreach my $n ($start .. $end) {
    my $semi = ($n == $end ? ";\n" : '');
    my ($x,$y) = $path->n_to_xy($n);
    my @n_list = $path->xy_to_n_list($x,$y);
    ($x,$y) = transform($x,$y);
    if ($n != $start && $n != $end
        && grep {($_ & $mask) != ($n & $mask)} @n_list) {
      my ($sdx,$sdy) = $path->n_to_dxdy($n-1);
      my ($tdx,$tdy) = $path->n_to_dxdy($n);
      ($sdx,$sdy) = transform_rotate($sdx,$sdy);
      ($tdx,$tdy) = transform_rotate($tdx,$tdy);
      my $sx = $x . chamfer(-$sdx);
      my $sy = $y . chamfer(-$sdy);
      my $tx = $x . chamfer($tdx);
      my $ty = $y . chamfer($tdy);
      printwrap("$join($sx,$sy)");
      printwrap("--($tx,$ty)$semi");
    } else {
      printwrap("$join($x,$y)$semi");
    }
    $join = '--';
  }
}

print "    % end generated\n";
exit 0;
