#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use Set::IntSpan::Fast;
use Math::PlanePath::DragonCurve;
use Math::PlanePath::Base::Digits 'round_down_pow';

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'xy_rotate_plus90';

# uncomment this to run the ### lines
# use Smart::Comments;


my $type = '1-inf';
if (@ARGV) { $type = $ARGV[0]; }

my $pred_count = 16;
my $at_count = 15;

my $path = Math::PlanePath::DragonCurve->new;

my $indent = "         ";

sub n_is_left_boundary {
  my ($n) = @_;
  ### path_n_is_left_boundary(): $n
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to left side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! defined $path->xyxy_to_n_either($x1,$y1, $x1s,$y1s)
          || ! defined $path->xyxy_to_n_either($x2,$y2, $x2s,$y2s)
          || ! defined $path->xyxy_to_n_either($x1s,$y1s, $x2s,$y2s));
}
sub n_is_left_boundary_any {
  my ($n) = @_;
  ### path_n_is_left_boundary(): $n
  my ($n_limit,$exp) = round_down_pow($n,2);
  $n_limit *= 2;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to left side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  foreach my $n_other ($path->xyxy_to_n_either($x1,$y1, $x1s,$y1s),
                       $path->xyxy_to_n_either($x2,$y2, $x2s,$y2s),
                       $path->xyxy_to_n_either($x1s,$y1s, $x2s,$y2s)) {
    if (! defined $n_other
        || $n_other >= $n_limit) {
      return 1;
    }
  }
  return 0;
}

sub set_count_runs {
  my ($set) = @_;
  my $count = 0;
  my $iter = $set->iterate_runs();
  while ($iter->()) { $count++; }
  return $count;
}
sub set_pari_string {
  my ($set) = @_;
  my @runs;
  my $iter = $set->iterate_runs();
  while (my ($from,$to) = $iter->()) {
    if ($from == $to) {
      push @runs, "[$from]";
    } else {
      push @runs, "[$from..$to]";
    }
  }
  return 'concat([' . join(',',@runs) . '])';
}

#------------------------------------------------------------------------------

print "% generated by tools/$FindBin::Script $type\n";

if ($type eq '1-inf') {
  my $set = Set::IntSpan::Fast->new;
  my $count = 0;
  for (my $n = 0; ; $n++) {
    if (n_is_left_boundary($n)) {
      $set->add($n);
      $count++;
    } else {
      last if set_count_runs($set) >= 4;
    }
  }
  my $iter = $set->iterate_runs();
  while (my ($from,$to) = $iter->()) {
    print $indent;
    if ($from+2 >= $to) {
      print join(',',$from..$to),", \\;\n";
    } else {
      print "$from \\MyTightDots $to, \\;\n";
    }
  }
  print $indent, "\\dotsc\n";
  print "% GP-DEFINE  LpredInf_one_at_samples = ",set_pari_string($set),";\n";
}

if ($type eq '0-inf') {
  my $set = Set::IntSpan::Fast->new;
  my $count = 0;
  for (my $n = 0; ; $n++) {
    if (n_is_left_boundary($n)) {
      last if set_count_runs($set) >= 4;
    } else {
      $set->add($n);
      $count++;
    }
  }
  my $iter = $set->iterate_runs();
  while (my ($from,$to) = $iter->()) {
    print $indent;
    if ($from+2 >= $to) {
      print join(',',$from..$to),", \\;\n";
    } else {
      print "$from \\MyTightDots$to, \\;\n";
    }
  }
  print $indent, "\\dotsc\n";
  print "% GP-DEFINE  LpredInf_zero_at_samples = ",set_pari_string($set),";\n";
}

if ($type eq '1-any') {
  my $set = Set::IntSpan::Fast->new;
  my $count = 0;
  for (my $n = 0; ; $n++) {
    if (n_is_left_boundary_any($n)) {
      $set->add($n);
      $count++;
    } else {
      last if set_count_runs($set) >= 4;
    }
  }
  my $iter = $set->iterate_runs();
  while (my ($from,$to) = $iter->()) {
    print $indent;
    if ($from+2 >= $to) {
      print join(',',$from..$to),", \\;\n";
    } else {
      print "$from \\MyTightDots $to, \\;\n";
    }
  }
  print $indent, "\\dotsc\n";
  print "% GP-DEFINE  LpredAny_one_at_samples = ",set_pari_string($set),";\n";
}

if ($type eq '0-any') {
  my $set = Set::IntSpan::Fast->new;
  my $count = 0;
  for (my $n = 0; ; $n++) {
    if (n_is_left_boundary_any($n)) {
      last if set_count_runs($set) >= 4;
    } else {
      $set->add($n);
      $count++;
    }
  }
  my $iter = $set->iterate_runs();
  while (my ($from,$to) = $iter->()) {
    print $indent;
    if ($from+2 >= $to) {
      print join(',',$from..$to),", \\;\n";
    } else {
      print "$from \\MyTightDots $to, \\;\n";
    }
  }
  print $indent, "\\dotsc\n";
  print "% GP-DEFINE  LpredAny_zero_at_samples = ",set_pari_string($set),";\n";
}

print "% end generated\n";

exit 0;
