#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use List::Util 'min';
use Math::PlanePath::DragonCurve 118; # v.118 for xyxy_to_n_either()
use Math::PlanePath::Base::Digits 'round_down_pow';

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'xy_rotate_plus90','@dir4_to_dx','@dir4_to_dy';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 7;
$Printwrap = 2;
printwrap("% generated by tools/$FindBin::Script\n");

my $path = Math::PlanePath::DragonCurve->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
print "  \\newcommand{\\MyLevel}{$level}  % N=$n_lo to $n_hi\n";

foreach my $part ('level','infinite','all') {
  $Printwrap = 2;
  printwrap("\\begin{scope}[shift=(pos-$part)]\n");
  $Printwrap += 2;

  {
    my @n_list = (0);
    foreach my $k ($part eq 'level' ? ($level)
                   : $part eq 'infinite' ? ()
                   : (1 .. $level)) {
      my ($n_lo, $n_hi) = $path->level_to_n_range($k);
      push @n_list, $n_hi;
    }
    foreach my $i (0 .. $#n_list) {
      my $n = $n_list[$i];
      my ($x,$y) = $path->n_to_xy($n);
      my $name = "p$i";
      if ($i <= 1) { $name = "$name-$part"; }
      printwrap("\\node ($name) at ($x,$y) [my dot];\n");
    }
  }

  my $n_top = ($part eq 'level' ? $n_hi - 1 : $n_hi);

  my $thin = MyLineSegments->new;
  my $thick = MyLineSegments->new;
  foreach my $n ($n_lo .. $n_top) {
    my $is_boundary;
    if ($part eq 'infinite') {
      $is_boundary = path_n_is_left_boundary($path, $n, 8*$n_hi);
    } elsif ($part eq 'level') {
      $is_boundary = path_n_is_left_boundary($path, $n, $n_hi);
    } else {
      my ($n_limit) = round_down_pow($n,2);
      $n_limit *= 2;
      $is_boundary = path_n_is_left_boundary($path, $n, $n_limit);
    }
    my $segments = ($is_boundary ? $thick : $thin);
    $segments->add($path->n_to_xy($n),
                   $path->n_to_xy($n+1));
  }

  my @n_extra_dashed;

  # thick continuation
  if ($part eq 'infinite' || $part eq 'all') {
    my $count = 0;
    for (my $n = $n_top; ; $n++) {
      if (path_n_is_left_boundary($path, $n, 8*$n_hi)) {
        $thick->add($path->n_to_xy($n),
                    $path->n_to_xy($n+1));
        last if ++$count == 5;
        push @n_extra_dashed, $n+1;
      }
    }
  }

  if ($part eq 'infinite' || $part eq 'all') {
    $Printwrap = 4;
    printwrap("\\draw[my thin]");
    $Printwrap += 2;
    foreach my $n ($n_lo .. $n_top, @n_extra_dashed) {
      my ($x,$y) = $path->n_to_xy($n);
      foreach my $dir (0 .. $#dir4_to_dx) {
        my $dx = $dir4_to_dx[$dir];
        my $dy = $dir4_to_dy[$dir];
        my $x2 = $x + $dx;
        my $y2 = $y + $dy;

        my $n_seg = $path->xyxy_to_n_either($x,$y, $x2,$y2) // next;
        next if $n_seg < $n_top;

        $x2 = $x + $dx*.75;
        $y2 = $y + $dy*.75;
        printwrap(" ($x,$y)--($x2,$y2)");
      }
    }
    printwrap(";\n");

  }
  # draw thick last so its width overlaps the thin lines where the two touch
  foreach my $thickness ('thin','thick') {
    my $segments = ($thickness eq 'thick' ? $thick : $thin);
    $Printwrap = 4;
    printwrap("\\draw[my $thickness] ");
    $Printwrap += 2;
    $segments->printwrap;
    printwrap(";\n");
  }

  $Printwrap = 2;
  printwrap("\\end{scope}\n");
}
printwrap("% end generated\n");
exit 0;


sub path_n_is_left_boundary {
  my ($path, $n, $n_limit) = @_;
  ### path_n_is_left_boundary(): $n
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit));
}
sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
  return defined $n && $n < $n_limit;
}
