#!/usr/bin/perl -w

# Copyright 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon ':all';
use MyLineSegments;
use MyPrintwrap ':all';

my $type = 'ER-to-OL';
if (@ARGV) { $type = shift @ARGV; }

print "  % generated by tools/$FindBin::Script $type\n";

my %dxdy = ("1,0" => 1,
            "-1,0" => 1,
            "0,1" => 1,
            "0,-1" => 1,

            "1,1" => 1,
            "1,-1" => 1,
            "-1,1" => 1,
            "-1,-1" => 1,
           );

my %type_to_elem =
  ('ER-to-OL' =>
   [
    [ [2,0], [3,0], [4,0], [5,0] ], # dots
    [ [1,0], [1,1], [6,0], [6,1] ], # blanks
    [ [0,1, 1,1],
      [1,0, 1,1],
      [1,0, 2,0],
      [2,0, 2,1], [2,0, 2,-1],
      [2,0, 3,0],
      [3,0, 3,1], [3,0, 3,-1],
      [3,0, 4,0],
      [4,0, 4,1], [4,0, 4,-1],
      [4,0, 5,0],
      [5,0, 5,1], [5,0, 5,-1],
      [5,0, 6,0],
      [6,0, 6,1], [6,1, 7,1],
    ],
   ],

   'EL-to-ER-diamonds' =>
   [
    [ [2,0], [4,0], [6,0],   # dots
    ],
    [ [0,0],[8,0],        # blanks
      [-1,1], [1,1], [3,1], [5,1], [7,1], [9,1],
      [1,-1], [3,-1], [5,-1], [7,-1],
    ],
    [ [0,0, -1,1], [0,0, 1,1],
      [2,0, 1,1], [2,0, 1,-1], [2,0, 3,1], [2,0, 3,-1],
      [4,0, 3,1], [4,0, 3,-1], [4,0, 5,1], [4,0, 5,-1],
      [6,0, 5,1], [6,0, 5,-1], [6,0, 7,1], [6,0, 7,-1],
      [8,0, 7,1], [8,0, 9,1],
    ],
   ],

   'OR-to-EL' =>
   [
    [ [1,0], [2,0], [3,0], [4,0],   # dots
    ],
    [ [0,0], [0,1],        # blanks
      [1,1], [2,1], [3,1], [4,1],
      [5,0], [5,1],
    ],
    [ [-1,1, 0,1],
      [0,0, 0,1],
      [0,0, 1,0],
      [1,0, 1,1], [1,1, 2,1],
      [1,0, 2,0],
      [2,0, 2,1],
      [2,0, 3,0],
      [3,0, 3,1], [3,1, 4,1],
      [3,0, 4,0],
      [4,0, 4,1],
      [4,0, 5,0],
      [5,0, 5,1], [5,1, 6,1],
    ],
   ],

   'OL-to-OR-diamonds' =>
   [
    [ [2,1], [4,1], [6,1],   # dots
    ],
    [ [0,1],[8,1],        # blanks
      [-1,2], [1,2], [3,2], [5,2], [7,2], [9,2],
      [1,0], [3,0], [5,0], [7,0],
    ],
    [ [0,1, -1,2], [0,1, 1,2],
      [2,1, 1,2], [2,1, 1,0], [2,1, 3,2], [2,1, 3,0],
      [4,1, 3,2], [4,1, 3,0], [4,1, 5,2], [4,1, 5,0],
      [6,1, 5,2], [6,1, 5,0], [6,1, 7,2], [6,1, 7,0],
      [8,1, 7,2], [8,1, 9,2],
    ],
   ],
  );

my $elem = $type_to_elem{$type}
  || die "unrecognised type $type";
my ($dots, $blanks, $segments) = @$elem;

foreach my $expand (0, 1) {
  my %points;

  $Printwrap = 2;
  if ($expand) {
    printwrap("\\begin{scope}[shift=(pos-right)]\n");
  }

  foreach my $dot (@$dots) {
    my ($x,$y) = @$dot;
    my $name = "p$x$y";
    if ($expand) { $name .= "-expand"; }
    printwrap(" \\node ($name) at ($x,$y) [my dot];");
    if ($points{"$x,$y"}) { die "duplicate $x,$y"; }
    $points{"$x,$y"} = $name;
  }
  printwrap_newline();

  foreach my $blank (@$blanks) {
    my ($x,$y) = @$blank;
    my $name = "p$x$y";
    if ($expand) { $name .= "-expand"; }
    printwrap(" \\node ($name) at ($x,$y) [my blank dot];");
    if ($points{"$x,$y"}) { die "duplicate"; }
    $points{"$x,$y"} = $name;
  }
  printwrap_newline();

  foreach my $segment (@$segments) {
    my ($x1,$y1, $x2,$y2) = @$segment;
    if (($x1+$y1) % 2) {
      ($x1,$y1, $x2,$y2) = ($x2,$y2, $x1,$y1);
    }
    my $dx = $x2 - $x1;
    my $dy = $y2 - $y1;
    $dxdy{"$dx,$dy"} or die;

    if ($expand) {
      printwrap(" \\pic at ($x1,$y1) [transform shape] {my expand pic={$dx,$dy}};");
    } else {
      my $from = $points{"$x1,$y1"} || "$x1,$y1";
      my $to   = $points{"$x2,$y2"} || "$x2,$y2";
      printwrap(" \\draw[->] ($from)--($to);");
    }
  }
  printwrap_newline();

  if ($expand) {
    printwrap("\\end{scope}\n");
  }
}

print "  % end generated\n";
exit 0;
