#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()
use Math::PlanePath::ComplexPlus 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon '@dir4_to_dx','@dir4_to_dy';
use MyLineSegments;
use MyPrintwrap ':all';


my $level = 4;
if (@ARGV) { $level = shift @ARGV; }

print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

# (x+y*I)/(1+I)*-I = (-1/2*x + 1/2*y + (- 1/2*x - 1/2*y)*I
#
sub rotate_minus135 {
  my ($x,$y) = @_;
  return (-$x+$y, -$x-$y);
}

{
  my $path = Math::PlanePath::DragonCurve->new;
  my ($n_lo, $n_hi) = $path->level_to_n_range($level+1);
  my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);

  {
    my ($x,$y) = rotate_minus135($x_hi,$y_hi);
    print "  \\coordinate (end) at ($x,$y);\n";
  }

  my $sign = sub {
    my ($x1,$x2) = @_;
    return ($x2 > $x1 ? '+' : '-');
  };

  print "  \\begin{scope}\n";
  $Printwrap = 4;
  foreach my $rot (0,1) {
    my ($prev_x,$prev_y);
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x1,$y1) = $path->n_to_xy($n);
      my ($x2,$y2) = $path->n_to_xy($n+1);
      if ($rot) {
        $x1 = $x_hi - $x1;
        $y1 = $y_hi - $y1;
        $x2 = $x_hi - $x2;
        $y2 = $y_hi - $y2;
      }
      ($x1,$y1) = rotate_minus135($x1,$y1);
      ($x2,$y2) = rotate_minus135($x2,$y2);

      my $prev = (defined $prev_x
                  ? "($prev_x,$prev_y)--"
                  : '');
      ($x1,$y1, $x2,$y2) = ("$x1" . $sign->($x1,$x2) . "\\MyGap",
                            "$y1" . $sign->($y1,$y2) . "\\MyGap",
                            "$x2" . $sign->($x2,$x1) . "\\MyGap",
                            "$y2" . $sign->($y2,$y1) . "\\MyGap");
      if ($n % 2) {
        if ($prev) { printwrap(" \\draw$prev($x1,$y1);"); }
        printwrap(" \\draw[<-]($x1,$y1)--($x2,$y2);");
      } else {
        printwrap(" \\draw[->]$prev($x1,$y1)--($x2,$y2);");
      }

      $prev_x = $x2;
      $prev_y = $y2;
    }
  }
  printwrap_newline();

  # # numbers
  # {
  #   foreach my $n ($n_lo .. $n_hi) {
  #     my ($x,$y) = $path->n_to_xy($n);
  #     $x = 2*$x;
  #     $y = 2*$y - 1;
  #     printwrap(" \\node at ($x,$y) {$n};");
  #   }
  #   printwrap_newline();
  # }

  print "  \\end{scope}\n";
}

{
  my $path = Math::PlanePath::ComplexPlus->new;
  my ($n_lo, $n_hi) = $path->level_to_n_range($level);

  my $insides = MyLineSegments->new;
  my $boundary = MyLineSegments->new;

  print "  \\begin{scope}[my right pos,font=\\tiny]\n";

  # numbers
  {
    foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      $x = 2*$x;
      $y = 2*$y - 1;
      printwrap(" \\node at ($x,$y) {$n};");
    }
    printwrap_newline();
  }

  {
    $Printwrap = 3;
    foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      foreach my $dir4 (0 .. 3) {
        my $dx = $dir4_to_dx[$dir4];
        my $dy = $dir4_to_dy[$dir4];
        my $n_other = $path->xy_to_n($x+$dx,$y+$dy);
        my $segs = (defined $n_other && $n_other <= $n_hi
                    ? $insides     # inside side
                    : $boundary);  # boundary side
        $segs->add(2*$x+$dx - $dy,  2*$y+$dy - $dx - 1,
                   2*$x+$dx + $dy,  2*$y+$dy + $dx - 1);
      }
    }

    $Printwrap = 4;
    printwrap("\\draw ");
    $Printwrap += 2;
    $boundary->printwrap;
    printwrap(";\n");

    $Printwrap = 4;
    printwrap("\\draw[dotted] ");
    $Printwrap += 2;
    $insides->printwrap;
    printwrap(";\n");

    # inside corners
    {
      my $is_visited = sub {
        my ($x,$y) = @_;
        my $n = $path->xy_to_n($x,$y);
        return (defined $n && $n <= $n_hi);
      };
      $Printwrap = 4;
      printwrap("\\path foreach \\p in {");
      $Printwrap = 6;
      foreach my $n ($n_lo .. $n_hi) {
        my ($x,$y) = $path->n_to_xy($n);
        if ($is_visited->($x+1,$y)
            && $is_visited->($x+1,$y+1)
            && $is_visited->($x,$y+1)) {
          $x = 2*$x + 1;
          $y = 2*$y;
          printwrap("($x,$y),");
        }
      }
      printwrap("}");
      printwrap(" { \\p node[my dot] };\n");
    }
  }

  printwrap_newline();
  print "  \\end{scope}\n";

  # /2 because 2*X,2*Y coordinates
  my $boundary_length = $boundary->length / 2;
  my $insides_length = $insides->length / 2;
  print "  % boundary length $boundary_length insides $insides_length\n";
  print "% GP-Test  TS($level) == $boundary_length\n";
  print "% GP-Test  TD($level) == $insides_length\n";
}
print "  % end generated\n";
exit 0;


__END__

# my $xyxy_is_twindragon_touch = sub {
#   my ($x,$y, $level) = @_;
#   ### xyxy_is_twindragon_touch(): "$x, $y"
#   {
#     my $n = $path->xy_to_n($x,$y);
#     ### first: $n
#     unless ($n && $n < $n_hi) {
#       ### bad n: $n_hi
#       return 0;
#     }
#   }
#   ($x,$y) = ($x_hi-$x, $y_hi-$y);
#   ### reverse: "$x, $y"
#   {
#     my $n = $path->xy_to_n($x,$y);
#     ### second: $n
#     unless ($n && $n < $n_hi) {
#       ### bad n: $n_hi
#       return 0;
#     }
#   }
#   return 1;
# };

# my $touch = $xyxy_is_twindragon_touch->($x2,$y2);
# if ($touch) {
#   # my $dx1 = $x2 - $x1;
#   # my $dy1 = $y2 - $y1;
#   # my $dx2 = $x3 - $x2;
#   # my $dy2 = $y3 - $y2;
#   # my $sx = "$x2-\\MyChamfer*$dx1";
#   # my $sy = "$y2-\\MyChamfer*$dy1";
#   # my $tx = "$x2+\\MyChamfer*$dx2";
#   # my $ty = "$y2+\\MyChamfer*$dy2";
#   # printwrap_newline();
#   # printwrap("\\draw[<->]($x1,$y1)");
#   # $Printwrap = 4;
#   # printwrap("--($sx,$sy)");
#   # printwrap("--($tx,$ty)");
#   # printwrap("--($x3,$y3);\n");
#   # $Printwrap = 2;
# } else {
